# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendList,
    PyLegendSequence,
    PyLegendUnion,
    PyLegendCallable,
    PyLegendOptional,
)
from pylegend.core.language import (
    PyLegendBoolean,
    PyLegendPrimitive,
    convert_literal_to_literal_expression,
)
from pylegend.core.language.legendql_api.legendql_api_tds_row import LegendQLApiTdsRow
from pylegend.core.language.shared.helpers import generate_pure_lambda
from pylegend.core.sql.metamodel import (
    QuerySpecification,
)
from pylegend.core.tds.legendql_api.frames.legendql_api_applied_function_tds_frame import LegendQLApiAppliedFunction
from pylegend.core.tds.legendql_api.frames.legendql_api_base_tds_frame import LegendQLApiBaseTdsFrame
from pylegend.core.tds.legendql_api.frames.legendql_api_tds_frame import LegendQLApiTdsFrame
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToPureConfig
from pylegend.core.tds.tds_frame import FrameToSqlConfig

__all__: PyLegendSequence[str] = [
    "LegendQLApiAsOfJoinFunction"
]


class LegendQLApiAsOfJoinFunction(LegendQLApiAppliedFunction):
    __base_frame: LegendQLApiBaseTdsFrame
    __other_frame: LegendQLApiBaseTdsFrame
    __match_function: PyLegendCallable[[LegendQLApiTdsRow, LegendQLApiTdsRow], PyLegendUnion[bool, PyLegendBoolean]]
    __join_condition: PyLegendOptional[
        PyLegendCallable[[LegendQLApiTdsRow, LegendQLApiTdsRow], PyLegendUnion[bool, PyLegendBoolean]]
    ]

    @classmethod
    def name(cls) -> str:
        return "asOfJoin"

    def __init__(
            self,
            base_frame: LegendQLApiBaseTdsFrame,
            other_frame: LegendQLApiTdsFrame,
            match_function: PyLegendCallable[[LegendQLApiTdsRow, LegendQLApiTdsRow], PyLegendUnion[bool, PyLegendBoolean]],
            join_condition: PyLegendOptional[
                PyLegendCallable[[LegendQLApiTdsRow, LegendQLApiTdsRow], PyLegendUnion[bool, PyLegendBoolean]]
            ] = None
    ) -> None:
        self.__base_frame = base_frame
        if not isinstance(other_frame, LegendQLApiBaseTdsFrame):
            raise ValueError("Expected LegendQLApiBaseTdsFrame")  # pragma: no cover
        self.__other_frame = other_frame
        self.__match_function = match_function
        self.__join_condition = join_condition

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        raise RuntimeError("AsOfJoin SQL translation not supported yet")

    def to_pure(self, config: FrameToPureConfig) -> str:
        left_row = LegendQLApiTdsRow.from_tds_frame("l", self.__base_frame)
        right_row = LegendQLApiTdsRow.from_tds_frame("r", self.__other_frame)
        match_expr = self.__match_function(left_row, right_row)
        match_string = (match_expr.to_pure_expression(config.push_indent(2))
                        if isinstance(match_expr, PyLegendPrimitive) else
                        convert_literal_to_literal_expression(match_expr).to_pure_expression(config.push_indent(2)))
        if self.__join_condition is not None:
            join_expr = self.__join_condition(left_row, right_row)
            join_string = (join_expr.to_pure_expression(config.push_indent(2))
                           if isinstance(join_expr, PyLegendPrimitive) else
                           convert_literal_to_literal_expression(join_expr).to_pure_expression(config.push_indent(2)))
            join_arg = (
                f",{config.separator(2, True)}"
                f"{generate_pure_lambda('l, r', join_string)}{config.separator(1)}"
            )
        else:
            join_arg = f"{config.separator(1)}"

        return (f"{self.__base_frame.to_pure(config)}{config.separator(1)}" +
                f"->asOfJoin({config.separator(2)}"
                f"{self.__other_frame.to_pure(config.push_indent(2))},{config.separator(2, True)}"
                f"{generate_pure_lambda('l, r', match_string)}"
                f"{join_arg}"
                f")")

    def base_frame(self) -> LegendQLApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["LegendQLApiBaseTdsFrame"]:
        return [self.__other_frame]

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        return (
            [c.copy() for c in self.__base_frame.columns()] +
            [c.copy() for c in self.__other_frame.columns()]
        )

    def validate(self) -> bool:
        copy = self.__match_function  # For MyPy
        if not isinstance(copy, type(lambda x: 0)) or (copy.__code__.co_argcount != 2):
            raise TypeError("AsOfJoin match function should be a lambda which takes two arguments (TDSRow, TDSRow)")

        left_row = LegendQLApiTdsRow.from_tds_frame("left", self.__base_frame)
        right_row = LegendQLApiTdsRow.from_tds_frame("right", self.__other_frame)

        try:
            result = self.__match_function(left_row, right_row)
        except Exception as e:
            raise RuntimeError(
                "AsOfJoin match function incompatible. Error occurred while evaluating. Message: " + str(e)
            ) from e

        if not isinstance(result, (bool, PyLegendBoolean)):
            raise RuntimeError("AsOfJoin match function incompatible. Returns non boolean - " + str(type(result)))

        if self.__join_condition is not None:
            copy = self.__join_condition  # For MyPy
            if not isinstance(copy, type(lambda x: 0)) or (copy.__code__.co_argcount != 2):
                raise TypeError("AsOfJoin join function should be a lambda which takes two arguments (TDSRow, TDSRow)")

            try:
                result = self.__join_condition(left_row, right_row)
            except Exception as e:
                raise RuntimeError(
                    "AsOfJoin join function incompatible. Error occurred while evaluating. Message: " + str(e)
                ) from e

            if not isinstance(result, (bool, PyLegendBoolean)):
                raise RuntimeError("AsOfJoin join function incompatible. Returns non boolean - " + str(type(result)))

        left_cols = [c.get_name() for c in self.__base_frame.columns()]
        right_cols = [c.get_name() for c in self.__other_frame.columns()]

        final_cols = left_cols + right_cols
        if len(final_cols) != len(set(final_cols)):
            raise ValueError(
                "Found duplicate columns in joined frames. Use rename function to ensure there are no duplicate columns "
                f"in joined frames. Columns - Left Frame: {left_cols}, Right Frame: {right_cols}"
            )

        return True



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendList,
    PyLegendSet,
    PyLegendSequence,
    PyLegendUnion,
    PyLegendOptional
)
from pylegend.core.language.shared.primitives.boolean import PyLegendBoolean
from pylegend.core.language.shared.primitives.integer import PyLegendInteger
from pylegend.core.sql.metamodel import (
    QuerySpecification
)
from pylegend.core.tds.pandas_api.frames.functions.filter import PandasApiFilterFunction
from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import PandasApiAppliedFunction
from pylegend.core.tds.pandas_api.frames.pandas_api_base_tds_frame import PandasApiBaseTdsFrame
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToPureConfig, FrameToSqlConfig

__all__: PyLegendSequence[str] = [
    "PandasApiDropFunction"
]


class PandasApiDropFunction(PandasApiAppliedFunction):
    __base_frame: PandasApiBaseTdsFrame
    __labels: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str], PyLegendSet[str]]]
    __axis: PyLegendUnion[str, int, PyLegendInteger]
    __index: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str], PyLegendSet[str]]]
    __columns: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str], PyLegendSet[str]]]
    __level: PyLegendOptional[PyLegendUnion[int, PyLegendInteger, str]]
    __inplace: PyLegendUnion[bool, PyLegendBoolean]
    __errors: str

    @classmethod
    def name(cls) -> str:
        return "drop"  # pragma: no cover

    def __init__(
            self,
            base_frame: PandasApiBaseTdsFrame,
            labels: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str], PyLegendSet[str]]],
            axis: PyLegendUnion[str, int, PyLegendInteger],
            index: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str], PyLegendSet[str]]],
            columns: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str], PyLegendSet[str]]],
            level: PyLegendOptional[PyLegendUnion[int, PyLegendInteger, str]],
            inplace: PyLegendUnion[bool, PyLegendBoolean],
            errors: str
    ) -> None:
        self.__base_frame = base_frame
        self.__labels = labels
        self.__axis = axis
        self.__index = index
        self.__columns = columns
        self.__level = level
        self.__inplace = inplace
        self.__errors = errors

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        base_cols = [c.get_name() for c in self.__base_frame.columns()]

        if self.__errors == "raise":
            not_found = [col for col in self.__columns if col not in base_cols]  # type: ignore
            if not_found:
                raise KeyError(f"{not_found} not found in axis")

        columns_to_retain = [col for col in base_cols if col not in self.__columns]  # type: ignore
        filter_func = PandasApiFilterFunction(
            base_frame=self.__base_frame,
            items=columns_to_retain,
            like=None,
            regex=None,
            axis=1
        )
        return filter_func.to_sql(config)

    def to_pure(self, config: FrameToPureConfig) -> str:
        base_cols = [c.get_name() for c in self.__base_frame.columns()]
        if self.__errors == "raise":
            not_found = [col for col in self.__columns if col not in base_cols]  # type: ignore
            if not_found:
                raise KeyError(f"{not_found} not found in axis")

        columns_to_retain = [col for col in base_cols if col not in self.__columns]  # type: ignore
        filter_func = PandasApiFilterFunction(
            base_frame=self.__base_frame,
            items=columns_to_retain,
            like=None,
            regex=None,
            axis=1
        )
        return filter_func.to_pure(config)

    def base_frame(self) -> PandasApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["PandasApiBaseTdsFrame"]:
        return []

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        base_cols = [c.copy() for c in self.__base_frame.columns()]
        if self.__columns is not None:
            new_cols = []
            for col in base_cols:
                if col.get_name() not in self.__columns:
                    new_cols.append(col.copy())
            return new_cols
        return base_cols  # pragma: no cover

    def validate(self) -> bool:
        valid_paramters: int = 0
        if self.__axis is not None:
            if isinstance(self.__axis, (str, int, PyLegendInteger)):
                if self.__axis != 1 and self.__axis != "columns":
                    if self.__axis == 0 or self.__axis == "index":
                        raise NotImplementedError(
                            f"Axis {self.__axis} is not supported for 'drop' function in PandasApi")
                    else:
                        raise ValueError(f"No axis named {self.__axis} for object type Tds DataFrame")
            else:
                raise TypeError(f"No axis named {self.__axis} for object type Tds DataFrame")  # pragma: no cover
        if self.__level is not None:
            raise NotImplementedError("'level' parameter is not supported for 'drop' function in PandasApi")

        if self.__index is not None:
            raise NotImplementedError("'index' parameter is not supported for 'drop' function in PandasApi")

        if self.__labels is not None:
            valid_paramters += 1

            if self.__columns is None:
                self.__columns = self.__labels
            else:
                raise ValueError("Cannot specify both 'labels' and 'columns'")

        if self.__columns is not None:
            def _normalize_columns(columns):  # type: ignore
                if columns is None:
                    return []  # pragma: no cover
                if isinstance(columns, str):
                    return [columns]
                if isinstance(columns, (PyLegendSequence, PyLegendSet)):
                    return list(columns)
                raise TypeError(f"Unsupported type for columns: {type(columns)}")

            valid_paramters += 1
            self.__columns = _normalize_columns(self.__columns)  # type: ignore

        if isinstance(self.__inplace, (bool, PyLegendBoolean)):
            if self.__inplace is True:
                raise NotImplementedError(f"Only inplace=False is supported. Got inplace={self.__inplace!r}")
        else:
            raise TypeError(f"Inplace must be False. Got inplace={self.__inplace!r}")  # pragma: no cover

        if valid_paramters == 0:
            raise ValueError("Need to specify at least one of 'labels' or 'columns'")

        return True



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import date, datetime

from pylegend._typing import (
    PyLegendSequence,
    PyLegendOptional,
    PyLegendList,
    PyLegendUnion,
    PyLegendDict
)
from pylegend.core.language import (
    convert_literal_to_literal_expression
)
from pylegend.core.language.pandas_api.pandas_api_tds_row import PandasApiTdsRow
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    SingleColumn, FunctionCall, QualifiedName,
)
from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
    PandasApiAppliedFunction,
)
from pylegend.core.tds.pandas_api.frames.pandas_api_base_tds_frame import (
    PandasApiBaseTdsFrame,
)
from pylegend.core.tds.sql_query_helpers import copy_query
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToSqlConfig, FrameToPureConfig

__all__: PyLegendSequence[str] = ["PandasApiFillnaFunction"]


class PandasApiFillnaFunction(PandasApiAppliedFunction):
    __base_frame: PandasApiBaseTdsFrame
    __value: PyLegendUnion[
        int, float, str, bool, date, datetime,
        PyLegendDict[str, PyLegendUnion[int, float, str, bool, date, datetime]]
    ]
    __axis: PyLegendOptional[PyLegendUnion[int, str]]
    __inplace: bool
    __limit: PyLegendOptional[int]

    @classmethod
    def name(cls) -> str:
        return "fillna"  # pragma: no cover

    def __init__(
            self,
            base_frame: PandasApiBaseTdsFrame,
            value: PyLegendUnion[
                int, float, str, bool, date, datetime,
                PyLegendDict[str, PyLegendUnion[int, float, str, bool, date, datetime]]
            ],
            axis: PyLegendOptional[PyLegendUnion[int, str]],
            inplace: bool,
            limit: PyLegendOptional[int]
    ) -> None:
        self.__base_frame = base_frame
        self.__value = value
        self.__axis = axis
        self.__inplace = inplace
        self.__limit = limit

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        base_query = self.__base_frame.to_sql_query_object(config)
        new_query = copy_query(base_query)

        tds_row = PandasApiTdsRow.from_tds_frame("c", self.__base_frame)
        db_extension = config.sql_to_string_generator().get_db_extension()
        select_items = []

        for col in self.__base_frame.columns():
            col_name = col.get_name()
            fill_value = self.__value if not isinstance(self.__value, dict) else self.__value.get(col_name)
            col_expr = tds_row[col_name]
            col_sql_expr = col_expr.to_sql_expression({"c": new_query}, config)

            if fill_value is not None:
                fill_expr = convert_literal_to_literal_expression(fill_value)
                fill_sql_expr = fill_expr.to_sql_expression({"c": new_query}, config)
                sql_expr = FunctionCall(
                    name=QualifiedName(parts=['coalesce']),
                    distinct=False,
                    arguments=[col_sql_expr, fill_sql_expr],
                    filter_=None,
                    window=None
                )
            else:
                sql_expr = col_sql_expr  # type: ignore

            select_items.append(SingleColumn(alias=db_extension.quote_identifier(col_name), expression=sql_expr))

        new_query.select.selectItems = select_items  # type: ignore
        return new_query

    def to_pure(self, config: FrameToPureConfig) -> str:
        base_pure = self.__base_frame.to_pure(config)
        projections = []

        for col in self.__base_frame.columns():
            col_name = col.get_name()
            fill_value = self.__value if not isinstance(self.__value, dict) else self.__value.get(col_name)

            if fill_value is not None:
                fill_expr = convert_literal_to_literal_expression(fill_value)
                fill_pure_expr = fill_expr.to_pure_expression(config)
                projections.append(f"'{col_name}':c|coalesce($c.{col_name}, {fill_pure_expr})")
            else:
                projections.append(f"'{col_name}':c|$c.{col_name}")

        projection_string = ", ".join(projections)
        return f"{base_pure}{config.separator(1)}->project(~[{projection_string}])"

    def base_frame(self) -> PandasApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["PandasApiBaseTdsFrame"]:
        return []

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        return [c.copy() for c in self.__base_frame.columns()]

    def validate(self) -> bool:
        if self.__value is None:
            raise ValueError("Must specify a fill 'value'")

        if not isinstance(self.__value, (int, float, str, bool, date, datetime, dict)):
            raise TypeError(f"'value' parameter must be a scalar or dict, but you passed a {type(self.__value)}")
        if isinstance(self.__value, dict):
            for k, v in self.__value.items():
                if not isinstance(k, str):
                    raise TypeError(
                        "All keys in 'value' dict must be strings representing column names, "
                        f"but found key of type {type(k)}"
                    )
                if not isinstance(v, (int, float, str, bool, date, datetime)):
                    raise TypeError(f"Non-scalar value of type {type(v)} passed for column '{k}' in 'value' parameter")

        if self.__axis not in (0, 1, "index", "columns"):
            raise ValueError(f"No axis named {self.__axis} for object type TdsFrame")
        if self.__axis in (1, "columns"):
            raise NotImplementedError("axis=1 is not supported yet in Pandas API fillna")

        if self.__inplace:
            raise NotImplementedError("inplace=True is not supported yet in Pandas API fillna")

        if self.__limit is not None:
            raise NotImplementedError("limit parameter is not supported yet in Pandas API fillna")

        return True



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta
from pylegend._typing import (
    PyLegendSequence,
)
from pylegend.core.tds.abstract.frames.input_tds_frame import (
    InputTdsFrame,
    ExecutableInputTdsFrame,
    NonExecutableInputTdsFrame,
)
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.legacy_api.frames.legacy_api_base_tds_frame import LegacyApiBaseTdsFrame
from pylegend.core.request.legend_client import LegendClient


__all__: PyLegendSequence[str] = [
    "LegacyApiExecutableInputTdsFrame",
    "LegacyApiNonExecutableInputTdsFrame",
    "LegacyApiInputTdsFrame"
]


class LegacyApiInputTdsFrame(LegacyApiBaseTdsFrame, InputTdsFrame, metaclass=ABCMeta):
    def __init__(self, columns: PyLegendSequence[TdsColumn]) -> None:
        LegacyApiBaseTdsFrame.__init__(self, columns=columns)
        InputTdsFrame.__init__(self, columns=columns)


class LegacyApiExecutableInputTdsFrame(LegacyApiInputTdsFrame, ExecutableInputTdsFrame, metaclass=ABCMeta):
    def __init__(self, legend_client: LegendClient, columns: PyLegendSequence[TdsColumn]) -> None:
        LegacyApiInputTdsFrame.__init__(self, columns=columns)
        ExecutableInputTdsFrame.__init__(self, legend_client=legend_client, columns=columns)


class LegacyApiNonExecutableInputTdsFrame(LegacyApiInputTdsFrame, NonExecutableInputTdsFrame, metaclass=ABCMeta):
    def __init__(self, columns: PyLegendSequence[TdsColumn]) -> None:
        LegacyApiInputTdsFrame.__init__(self, columns=columns)
        NonExecutableInputTdsFrame.__init__(self, columns=columns)



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
from enum import Enum
import json
from pylegend._typing import (
    PyLegendList,
    PyLegendSequence
)

__all__: PyLegendSequence[str] = [
    "TdsColumn",
    "PrimitiveTdsColumn",
    "PrimitiveType",
    "EnumTdsColumn",
    "tds_columns_from_json",
]


class TdsColumn(metaclass=ABCMeta):
    __name: str

    def __init__(self, name: str) -> None:
        self.__name = name

    def get_name(self) -> str:
        return self.__name

    @abstractmethod
    def get_type(self) -> str:
        pass  # pragma: no cover

    @abstractmethod
    def copy(self) -> "TdsColumn":
        pass  # pragma: no cover

    @abstractmethod
    def copy_with_changed_name(self, new_name: str) -> "TdsColumn":
        pass  # pragma: no cover

    def __str__(self) -> str:
        return f"TdsColumn(Name: {self.get_name()}, Type: {self.get_type()})"


class PrimitiveType(Enum):
    Boolean = 1
    StrictDate = 2
    Number = 3
    String = 4
    LatestDate = 5
    Float = 6
    DateTime = 7
    Date = 8
    Integer = 9
    Decimal = 10


class PrimitiveTdsColumn(TdsColumn):
    __type: "PrimitiveType"

    def __init__(self, name: str, _type: "PrimitiveType") -> None:
        super().__init__(name)
        self.__type = _type

    def get_type(self) -> "str":
        return self.__type.name

    def copy(self) -> "TdsColumn":
        return PrimitiveTdsColumn(self.get_name(), self.__type)

    def copy_with_changed_name(self, new_name: str) -> "TdsColumn":
        return PrimitiveTdsColumn(new_name, self.__type)

    @classmethod
    def integer_column(cls, name: str) -> "PrimitiveTdsColumn":
        return PrimitiveTdsColumn(name, PrimitiveType.Integer)

    @classmethod
    def float_column(cls, name: str) -> "PrimitiveTdsColumn":
        return PrimitiveTdsColumn(name, PrimitiveType.Float)

    @classmethod
    def string_column(cls, name: str) -> "PrimitiveTdsColumn":
        return PrimitiveTdsColumn(name, PrimitiveType.String)

    @classmethod
    def boolean_column(cls, name: str) -> "PrimitiveTdsColumn":
        return PrimitiveTdsColumn(name, PrimitiveType.Boolean)

    @classmethod
    def number_column(cls, name: str) -> "PrimitiveTdsColumn":
        return PrimitiveTdsColumn(name, PrimitiveType.Number)

    @classmethod
    def date_column(cls, name: str) -> "PrimitiveTdsColumn":
        return PrimitiveTdsColumn(name, PrimitiveType.Date)

    @classmethod
    def datetime_column(cls, name: str) -> "PrimitiveTdsColumn":
        return PrimitiveTdsColumn(name, PrimitiveType.DateTime)

    @classmethod
    def strictdate_column(cls, name: str) -> "PrimitiveTdsColumn":
        return PrimitiveTdsColumn(name, PrimitiveType.StrictDate)


class EnumTdsColumn(TdsColumn):
    __enum_type: str
    __enum_values: PyLegendList[str]

    def __init__(self, name: str, enum_type: str, enum_values: PyLegendList[str]) -> None:
        super().__init__(name)
        self.__enum_type = enum_type
        self.__enum_values = enum_values

    def get_type(self) -> str:
        return self.__enum_type

    def copy(self) -> "TdsColumn":
        return EnumTdsColumn(self.get_name(), self.__enum_type, self.__enum_values)

    def copy_with_changed_name(self, new_name: str) -> "TdsColumn":
        return EnumTdsColumn(new_name, self.__enum_type, self.__enum_values)

    def get_enum_values(self) -> PyLegendList[str]:
        return self.__enum_values


def tds_columns_from_json(s: str) -> PyLegendSequence[TdsColumn]:
    try:
        parsed = json.loads(s)

        enums = parsed["enums"] if "enums" in parsed else []
        enums = [enums] if isinstance(enums, dict) else enums

        columns = parsed["columns"]
        columns = [columns] if isinstance(columns, dict) else columns

        result_columns: PyLegendList[TdsColumn] = []
        for col in columns:
            if col["_type"] == "primitiveSchemaColumn":
                result_columns.append(PrimitiveTdsColumn(col["name"], PrimitiveType[col["type"]]))
            else:
                result_columns.append(
                    EnumTdsColumn(col["name"], col["type"], _enum_values_for_type(col["type"], enums))
                )
        return result_columns

    except Exception as e:
        raise RuntimeError("Unable to parse tds columns from schema: \n" + s, e)


def _enum_values_for_type(enum_type: str, all_enums: PyLegendList[dict]) -> PyLegendList[str]:  # type: ignore
    for e in all_enums:
        if e["type"] == enum_type:
            values = e["values"]
            return values if isinstance(values, list) else [values]

    raise RuntimeError("Unknown enum type: " + enum_type)



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
)
from pylegend.core.language.shared.tds_row import AbstractTdsRow
from pylegend.core.tds.tds_frame import PyLegendTdsFrame

__all__: PyLegendSequence[str] = [
    "LegacyApiTdsRow",
]


class LegacyApiTdsRow(AbstractTdsRow):
    def __init__(self, frame_name: str, frame: PyLegendTdsFrame) -> None:
        super().__init__(frame_name, frame)

    @staticmethod
    def from_tds_frame(frame_name: str, frame: PyLegendTdsFrame) -> "LegacyApiTdsRow":
        return LegacyApiTdsRow(frame_name=frame_name, frame=frame)



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
)
from pylegend.legacy_api_tds_client import (
    LegacyApiTdsClient,
    legacy_api_tds_client,
)
from pylegend.legendql_api_tds_client import (
    LegendQLApiTdsClient,
    legendql_api_tds_client,
)
from pylegend.core.request import (
    LegendClient,
    AuthScheme,
    LocalhostEmptyAuthScheme,
    HeaderTokenAuthScheme,
    CookieAuthScheme,
    ResponseReader,
)
from pylegend.core.project_cooridnates import (
    VersionedProjectCoordinates,
    PersonalWorkspaceProjectCoordinates,
    GroupWorkspaceProjectCoordinates,
)
from pylegend.core.language import (
    agg,
    now,
    today,
    current_user,
)


__all__: PyLegendSequence[str] = [
    "__version__",

    "LegendQLApiTdsClient",
    "legendql_api_tds_client",

    "LegacyApiTdsClient",
    "legacy_api_tds_client",

    "LegendClient",
    "AuthScheme",
    "LocalhostEmptyAuthScheme",
    "HeaderTokenAuthScheme",
    "CookieAuthScheme",
    "ResponseReader",

    "VersionedProjectCoordinates",
    "PersonalWorkspaceProjectCoordinates",
    "GroupWorkspaceProjectCoordinates",

    "agg",
    "now",
    "today",
    "current_user",
]


import importlib.metadata

__version__ = importlib.metadata.version("pylegend")



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendList,
    PyLegendSequence,
    PyLegendCallable,
    PyLegendUnion,
)
from pylegend.core.language.legendql_api.legendql_api_custom_expressions import (
    LegendQLApiPrimitive,
    LegendQLApiSortInfo,
)
from pylegend.core.language.legendql_api.legendql_api_tds_row import LegendQLApiTdsRow
from pylegend.core.sql.metamodel import (
    QuerySpecification,
)
from pylegend.core.tds.legendql_api.frames.functions.legendql_api_function_helpers import infer_sorts_from_frame
from pylegend.core.tds.legendql_api.frames.legendql_api_applied_function_tds_frame import LegendQLApiAppliedFunction
from pylegend.core.tds.legendql_api.frames.legendql_api_base_tds_frame import LegendQLApiBaseTdsFrame
from pylegend.core.tds.sql_query_helpers import copy_query, create_sub_query
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToPureConfig
from pylegend.core.tds.tds_frame import FrameToSqlConfig

__all__: PyLegendSequence[str] = [
    "LegendQLApiSortFunction"
]


class LegendQLApiSortFunction(LegendQLApiAppliedFunction):
    __base_frame: LegendQLApiBaseTdsFrame
    __sort_infos: PyLegendList[LegendQLApiSortInfo]

    @classmethod
    def name(cls) -> str:
        return "sort"

    def __init__(
            self,
            base_frame: LegendQLApiBaseTdsFrame,
            sort_infos: PyLegendUnion[
                str,
                PyLegendList[str],
                PyLegendCallable[
                    [LegendQLApiTdsRow],
                    PyLegendUnion[
                        LegendQLApiPrimitive,
                        LegendQLApiSortInfo,
                        PyLegendList[PyLegendUnion[LegendQLApiPrimitive, LegendQLApiSortInfo]],
                    ]
                ]
            ]
    ) -> None:
        self.__base_frame = base_frame
        self.__sort_infos = infer_sorts_from_frame(base_frame, sort_infos, "'sort' function sort_infos")

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        base_query = self.__base_frame.to_sql_query_object(config)
        should_create_sub_query = (base_query.offset is not None) or (base_query.limit is not None)
        new_query = (
            create_sub_query(base_query, config, "root") if should_create_sub_query else
            copy_query(base_query)
        )
        new_query.orderBy = [i.to_sql_node(query=new_query, config=config) for i in self.__sort_infos]
        return new_query

    def to_pure(self, config: FrameToPureConfig) -> str:
        return (f"{self.__base_frame.to_pure(config)}{config.separator(1)}" +
                f"->sort([{', '.join([i.to_pure_expression(config) for i in self.__sort_infos])}])")

    def base_frame(self) -> LegendQLApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["LegendQLApiBaseTdsFrame"]:
        return []

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        return [c.copy() for c in self.__base_frame.columns()]

    def validate(self) -> bool:
        return True



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendList,
    PyLegendSequence,
    PyLegendUnion,
    PyLegendOptional,
    PyLegendCallable,
    PyLegendDict
)
from pylegend.core.language import (
    PyLegendInteger,
    PyLegendBoolean
)
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    SelectItem,
    SingleColumn
)
from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import PandasApiAppliedFunction
from pylegend.core.tds.pandas_api.frames.pandas_api_base_tds_frame import PandasApiBaseTdsFrame
from pylegend.core.tds.sql_query_helpers import copy_query
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToPureConfig, FrameToSqlConfig


class PandasApiRenameFunction(PandasApiAppliedFunction):
    __base_frame: PandasApiBaseTdsFrame
    __mapper: PyLegendOptional[PyLegendUnion[PyLegendDict[str, str], PyLegendCallable[[str], str]]]
    __axis: PyLegendUnion[str, int, PyLegendInteger]
    __index: PyLegendOptional[PyLegendUnion[PyLegendDict[str, str], PyLegendCallable[[str], str]]]
    __columns: PyLegendOptional[PyLegendUnion[PyLegendDict[str, str], PyLegendCallable[[str], str]]]
    __level: PyLegendOptional[PyLegendUnion[int, PyLegendInteger, str]]
    __inplace: PyLegendUnion[bool, PyLegendBoolean]
    __copy: PyLegendUnion[bool, PyLegendBoolean]
    __errors: str

    @classmethod
    def name(cls) -> str:
        return "rename"  # pragma: no cover

    def __init__(
            self,
            base_frame: PandasApiBaseTdsFrame,
            mapper: PyLegendOptional[PyLegendUnion[PyLegendDict[str, str], PyLegendCallable[[str], str]]],
            axis: PyLegendUnion[str, int, PyLegendInteger],
            index: PyLegendOptional[PyLegendUnion[PyLegendDict[str, str], PyLegendCallable[[str], str]]],
            columns: PyLegendOptional[PyLegendUnion[PyLegendDict[str, str], PyLegendCallable[[str], str]]],
            level: PyLegendOptional[PyLegendUnion[int, PyLegendInteger, str]],
            inplace: PyLegendUnion[bool, PyLegendBoolean],
            errors: str,
            copy: PyLegendUnion[bool, PyLegendBoolean]
    ) -> None:
        self.__base_frame = base_frame
        self.__mapper = mapper
        self.__axis = axis
        self.__index = index
        self.__columns = columns
        self.__level = level
        self.__inplace = inplace
        self.__errors = errors
        self.__copy = copy

    def __resolve_columns_mapping(self) -> PyLegendDict[str, str]:
        base_cols = [c.get_name() for c in self.__base_frame.columns()]
        mapping_source = None

        axis_is_columns = (self.__axis == 1 or self.__axis == "columns")

        # Priority: explicit columns=, else mapper when axis targets columns
        if self.__columns is not None:
            mapping_source = self.__columns
        elif self.__mapper is not None and axis_is_columns:
            mapping_source = self.__mapper

        if mapping_source is None:
            return {}

        if not callable(mapping_source) and not isinstance(mapping_source, dict):
            raise TypeError(
                f"Rename mapping must be a dict or a callable, got {type(mapping_source)}"
            )

        out: PyLegendDict[str, str] = {}
        if callable(mapping_source):
            func = mapping_source
            for col in base_cols:
                new = func(col)
                if not isinstance(new, str):
                    raise TypeError(
                        f"Rename function must return str, got {type(new)} for column {col}")  # pragma: no cover
                if new != col:
                    out[col] = new
        else:
            # dict-like
            dict_map: PyLegendDict[str, str] = mapping_source
            if self.__errors == "raise":
                missing = [k for k in dict_map.keys() if k not in base_cols]
                if missing:
                    raise KeyError(f"{missing} not found in axis")

            for k, v in dict_map.items():
                if k in base_cols and k != v:
                    out[k] = v

        return out

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        rename_map = self.__resolve_columns_mapping()
        base_query = self.__base_frame.to_sql_query_object(config)
        db_extension = config.sql_to_string_generator().get_db_extension()

        # Prepare quoted lookup for aliases
        quoted_from = [db_extension.quote_identifier(s) for s in rename_map.keys()]
        quoted_to = [db_extension.quote_identifier(rename_map[s]) for s in rename_map.keys()]

        new_select_items: PyLegendList[SelectItem] = []
        for col in base_query.select.selectItems:
            if not isinstance(col, SingleColumn):
                raise ValueError("Rename operation not supported for non-SingleColumn select items")  # pragma: no cover
            if col.alias is None:
                raise ValueError("Rename operation requires SingleColumn items with aliases")  # pragma: no cover
            if col.alias in quoted_from:
                new_alias = quoted_to[quoted_from.index(col.alias)]
                new_select_items.append(SingleColumn(alias=new_alias, expression=col.expression))
            else:
                new_select_items.append(col)

        new_query = copy_query(base_query)
        new_query.select.selectItems = new_select_items
        return new_query

    def to_pure(self, config: FrameToPureConfig) -> str:
        rename_map = self.__resolve_columns_mapping()
        base_pure = self.__base_frame.to_pure(config)

        # Build a single project that aliases columns to new names
        project_items: PyLegendList[str] = []
        for c in self.__base_frame.columns():
            orig = c.get_name()
            new = rename_map.get(orig, orig)
            project_items.append(f"{new}:x|$x.{orig}")

        project_body = ", ".join(project_items)
        return (
            f"{base_pure}{config.separator(1)}"
            f"->project({config.separator(2)}~[{project_body}]{config.separator(1)})"
        )

    def base_frame(self) -> PandasApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["PandasApiBaseTdsFrame"]:
        return []

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        rename_map = self.__resolve_columns_mapping()
        new_cols = []
        for c in self.__base_frame.columns():
            name = c.get_name()
            if name in rename_map:
                new_cols.append(c.copy_with_changed_name(rename_map[name]))
            else:
                new_cols.append(c.copy())
        names = [c.get_name() for c in new_cols]
        if len(names) != len(set(names)):
            raise ValueError("Resulting columns contain duplicates after rename")
        return new_cols

    def validate(self) -> bool:
        if self.__level is not None:
            raise NotImplementedError("level parameter not supported yet in Pandas API")

        if not isinstance(self.__inplace, bool):
            raise TypeError(f"inplace must be bool. Got {type(self.__inplace)}")
        if self.__inplace is True:
            raise NotImplementedError("inplace=True not supported yet in Pandas API")

        if not isinstance(self.__copy, bool):
            raise TypeError(f"copy must be bool. Got {type(self.__copy)}")
        if self.__copy is False:
            raise NotImplementedError("copy=False not supported yet in Pandas API")

        if self.__errors not in ("ignore", "raise"):
            raise ValueError(f"errors must be 'ignore' or 'raise'. Got {self.__errors}")

        # axis validation
        if self.__axis not in (1, "columns", 0, "index"):
            raise ValueError(f"Unsupported axis {self.__axis}")
        if self.__axis in (0, "index"):
            raise NotImplementedError("Renaming index not supported yet in Pandas API")

        # index
        if self.__index is not None:
            raise NotImplementedError("Index mapper not supported yet in Pandas API")

        # conflict validation
        if self.__mapper and self.__columns:
            raise ValueError("Cannot specify both 'axis' and any of 'index' or 'columns'")

        self.__resolve_columns_mapping()  # runs validation
        return True



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
from pylegend._typing import (
    PyLegendSequence,
    PyLegendGeneric,
    PyLegendTypeVar,
    TYPE_CHECKING
)
from pylegend.core.request.response_reader import ResponseReader
if TYPE_CHECKING:
    from pylegend.core.tds.tds_frame import PyLegendTdsFrame  # pragma: no cover

__all__: PyLegendSequence[str] = [
    "ResultHandler"
]

R = PyLegendTypeVar('R')


class ResultHandler(PyLegendGeneric[R], metaclass=ABCMeta):

    @abstractmethod
    def handle_result(self, frame: "PyLegendTdsFrame", result: ResponseReader) -> R:
        pass  # pragma: no cover



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
)
from pylegend.core.tds.result_handler.result_handler import ResultHandler
from pylegend.core.tds.result_handler.to_string_result_handler import ToStringResultHandler
from pylegend.core.tds.result_handler.to_json_file_result_handler import ToJsonFileResultHandler
from pylegend.core.tds.result_handler.to_csv_file_result_handler import ToCsvFileResultHandler

__all__: PyLegendSequence[str] = [
    "ResultHandler",
    "ToStringResultHandler",
    "ToJsonFileResultHandler",
    "ToCsvFileResultHandler",
]



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
    PyLegendUnion,
    TYPE_CHECKING,
)
from pylegend.core.language.shared.primitives.number import PyLegendNumber
from pylegend.core.language.shared.expression import PyLegendExpressionIntegerReturn
from pylegend.core.language.shared.literal_expressions import PyLegendIntegerLiteralExpression
from pylegend.core.sql.metamodel import (
    Expression,
    QuerySpecification
)
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.language.shared.operations.integer_operation_expressions import (
    PyLegendIntegerAddExpression,
    PyLegendIntegerAbsoluteExpression,
    PyLegendIntegerNegativeExpression,
    PyLegendIntegerSubtractExpression,
    PyLegendIntegerMultiplyExpression,
    PyLegendIntegerModuloExpression,
    PyLegendIntegerCharExpression
)
if TYPE_CHECKING:
    from pylegend.core.language.shared.primitives import PyLegendFloat
    from pylegend.core.language.shared.primitives.string import PyLegendString

__all__: PyLegendSequence[str] = [
    "PyLegendInteger"
]


class PyLegendInteger(PyLegendNumber):
    __value_copy: PyLegendExpressionIntegerReturn

    def __init__(
            self,
            value: PyLegendExpressionIntegerReturn
    ) -> None:
        self.__value_copy = value
        super().__init__(value)

    def char(self) -> "PyLegendString":
        from pylegend.core.language.shared.primitives.string import PyLegendString
        return PyLegendString(PyLegendIntegerCharExpression(self.__value_copy))

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return super().to_sql_expression(frame_name_to_base_query_map, config)

    def value(self) -> PyLegendExpressionIntegerReturn:
        return self.__value_copy

    def __add__(
            self,
            other: PyLegendUnion[int, float, "PyLegendInteger", "PyLegendFloat", "PyLegendNumber"]
    ) -> "PyLegendUnion[PyLegendNumber, PyLegendInteger]":
        PyLegendNumber.validate_param_to_be_number(other, "Integer plus (+) parameter")
        if isinstance(other, (int, PyLegendInteger)):
            other_op = PyLegendInteger.__convert_to_integer_expr(other)
            return PyLegendInteger(PyLegendIntegerAddExpression(self.__value_copy, other_op))
        else:
            return super().__add__(other)

    def __radd__(
            self,
            other: PyLegendUnion[int, float, "PyLegendInteger", "PyLegendFloat", "PyLegendNumber"]
    ) -> "PyLegendUnion[PyLegendNumber, PyLegendInteger]":
        PyLegendNumber.validate_param_to_be_number(other, "Integer plus (+) parameter")
        if isinstance(other, (int, PyLegendInteger)):
            other_op = PyLegendInteger.__convert_to_integer_expr(other)
            return PyLegendInteger(PyLegendIntegerAddExpression(other_op, self.__value_copy))
        else:
            return super().__radd__(other)

    def __sub__(
            self,
            other: PyLegendUnion[int, float, "PyLegendInteger", "PyLegendFloat", "PyLegendNumber"]
    ) -> "PyLegendUnion[PyLegendNumber, PyLegendInteger]":
        PyLegendNumber.validate_param_to_be_number(other, "Integer minus (-) parameter")
        if isinstance(other, (int, PyLegendInteger)):
            other_op = PyLegendInteger.__convert_to_integer_expr(other)
            return PyLegendInteger(PyLegendIntegerSubtractExpression(self.__value_copy, other_op))
        else:
            return super().__sub__(other)

    def __rsub__(
            self,
            other: PyLegendUnion[int, float, "PyLegendInteger", "PyLegendFloat", "PyLegendNumber"]
    ) -> "PyLegendUnion[PyLegendNumber, PyLegendInteger]":
        PyLegendNumber.validate_param_to_be_number(other, "Integer minus (-) parameter")
        if isinstance(other, (int, PyLegendInteger)):
            other_op = PyLegendInteger.__convert_to_integer_expr(other)
            return PyLegendInteger(PyLegendIntegerSubtractExpression(other_op, self.__value_copy))
        else:
            return super().__rsub__(other)

    def __mul__(
            self,
            other: PyLegendUnion[int, float, "PyLegendInteger", "PyLegendFloat", "PyLegendNumber"]
    ) -> "PyLegendUnion[PyLegendNumber, PyLegendInteger]":
        PyLegendNumber.validate_param_to_be_number(other, "Integer multiply (*) parameter")
        if isinstance(other, (int, PyLegendInteger)):
            other_op = PyLegendInteger.__convert_to_integer_expr(other)
            return PyLegendInteger(PyLegendIntegerMultiplyExpression(self.__value_copy, other_op))
        else:
            return super().__mul__(other)

    def __rmul__(
            self,
            other: PyLegendUnion[int, float, "PyLegendInteger", "PyLegendFloat", "PyLegendNumber"]
    ) -> "PyLegendUnion[PyLegendNumber, PyLegendInteger]":
        PyLegendNumber.validate_param_to_be_number(other, "Integer multiply (*) parameter")
        if isinstance(other, (int, PyLegendInteger)):
            other_op = PyLegendInteger.__convert_to_integer_expr(other)
            return PyLegendInteger(PyLegendIntegerMultiplyExpression(other_op, self.__value_copy))
        else:
            return super().__rmul__(other)

    def __mod__(
            self,
            other: PyLegendUnion[int, "PyLegendInteger"]
    ) -> "PyLegendUnion[PyLegendNumber, PyLegendInteger]":
        PyLegendInteger.__validate__param_to_be_integer(other, "Integer modulo (%) parameter")
        other_op = PyLegendInteger.__convert_to_integer_expr(other)
        return PyLegendInteger(PyLegendIntegerModuloExpression(self.__value_copy, other_op))

    def __rmod__(
            self,
            other: PyLegendUnion[int, "PyLegendInteger"]
    ) -> "PyLegendUnion[PyLegendNumber, PyLegendInteger]":
        PyLegendNumber.validate_param_to_be_number(other, "Integer modulo (%) parameter")
        other_op = PyLegendInteger.__convert_to_integer_expr(other)
        return PyLegendInteger(PyLegendIntegerModuloExpression(other_op, self.__value_copy))

    def __abs__(self) -> "PyLegendInteger":
        return PyLegendInteger(PyLegendIntegerAbsoluteExpression(self.__value_copy))

    def __neg__(self) -> "PyLegendInteger":
        return PyLegendInteger(PyLegendIntegerNegativeExpression(self.__value_copy))

    def __pos__(self) -> "PyLegendInteger":
        return self

    @staticmethod
    def __convert_to_integer_expr(
            val: PyLegendUnion[int, "PyLegendInteger"]
    ) -> PyLegendExpressionIntegerReturn:
        if isinstance(val, int):
            return PyLegendIntegerLiteralExpression(val)
        return val.__value_copy

    @staticmethod
    def __validate__param_to_be_integer(param: PyLegendUnion[int, "PyLegendInteger"], desc: str) -> None:
        if not isinstance(param, (int, PyLegendInteger)):
            raise TypeError(desc + " should be a int or an integer expression (PyLegendInteger)."
                                   " Got value " + str(param) + " of type: " + str(type(param)))



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
from pylegend._typing import (
    PyLegendSequence,
)
from pylegend.core.sql.metamodel import QuerySpecification
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.tds_frame import FrameToPureConfig
from pylegend.core.tds.abstract.frames.base_tds_frame import BaseTdsFrame


__all__: PyLegendSequence[str] = [
    "AppliedFunctionTdsFrame",
    "AppliedFunction",
]


class AppliedFunction(metaclass=ABCMeta):
    @classmethod
    @abstractmethod
    def name(cls) -> str:
        pass  # pragma: no cover

    @abstractmethod
    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        pass  # pragma: no cover

    def to_pure(self, config: FrameToPureConfig) -> str:
        raise RuntimeError("PURE generation not supported for '" + self.name() + "' function")

    @abstractmethod
    def base_frame(self) -> BaseTdsFrame:
        pass  # pragma: no cover

    @abstractmethod
    def tds_frame_parameters(self) -> PyLegendSequence[BaseTdsFrame]:
        pass  # pragma: no cover

    @abstractmethod
    def calculate_columns(self) -> PyLegendSequence[TdsColumn]:
        pass  # pragma: no cover

    @abstractmethod
    def validate(self) -> bool:
        pass  # pragma: no cover


class AppliedFunctionTdsFrame(BaseTdsFrame, metaclass=ABCMeta):
    __applied_function: AppliedFunction

    def __init__(self, applied_function: AppliedFunction):
        applied_function.validate()
        super().__init__(columns=applied_function.calculate_columns())
        self.__applied_function = applied_function

    def to_sql_query_object(self, config: FrameToSqlConfig) -> QuerySpecification:
        return self.__applied_function.to_sql(config)

    def to_pure(self, config: FrameToPureConfig) -> str:
        return self.__applied_function.to_pure(config)

    def get_all_tds_frames(self) -> PyLegendSequence[BaseTdsFrame]:
        return [
            y
            for x in [self.__applied_function.base_frame()] + list(self.__applied_function.tds_frame_parameters())
            for y in x.get_all_tds_frames()
        ] + [self]



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendList,
    PyLegendSequence,
    PyLegendCallable,
    PyLegendUnion,
    PyLegendTuple,
)
from pylegend.core.language import PyLegendColumnExpression
from pylegend.core.language.legendql_api.legendql_api_custom_expressions import LegendQLApiPrimitive
from pylegend.core.language.legendql_api.legendql_api_tds_row import LegendQLApiTdsRow
from pylegend.core.tds.legendql_api.frames.legendql_api_applied_function_tds_frame import LegendQLApiAppliedFunction
from pylegend.core.tds.sql_query_helpers import copy_query
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    SingleColumn,
    SelectItem,
)
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.tds_frame import FrameToPureConfig
from pylegend.core.tds.legendql_api.frames.legendql_api_base_tds_frame import LegendQLApiBaseTdsFrame
from pylegend.core.language.shared.helpers import escape_column_name


__all__: PyLegendSequence[str] = [
    "LegendQLApiRenameFunction"
]


class LegendQLApiRenameFunction(LegendQLApiAppliedFunction):
    __base_frame: LegendQLApiBaseTdsFrame
    __column_names: PyLegendList[str]
    __renamed_column_names: PyLegendList[str]

    @classmethod
    def name(cls) -> str:
        return "renameColumns"

    def __init__(
            self,
            base_frame: LegendQLApiBaseTdsFrame,
            column_renames: PyLegendUnion[
                PyLegendTuple[str, str],
                PyLegendList[PyLegendTuple[str, str]],
                PyLegendCallable[
                    [LegendQLApiTdsRow],
                    PyLegendUnion[
                        PyLegendTuple[LegendQLApiPrimitive, str],
                        PyLegendList[PyLegendTuple[LegendQLApiPrimitive, str]]
                    ]
                ]
            ]
    ) -> None:
        self.__base_frame = base_frame

        col_names: PyLegendList[str] = []
        renamed_col_names: PyLegendList[str] = []

        def rename_tuple_check(t):  # type: ignore
            return isinstance(t, tuple) and (len(t) == 2) and isinstance(t[0], str) and isinstance(t[1], str)

        if rename_tuple_check(column_renames):  # type: ignore
            col_names.append(column_renames[0])  # type: ignore
            renamed_col_names.append(column_renames[1])  # type: ignore

        elif isinstance(column_renames, list) and all([rename_tuple_check(r) for r in column_renames]):  # type: ignore
            for t in column_renames:
                col_names.append(t[0])
                renamed_col_names.append(t[1])

        elif isinstance(column_renames, type(lambda x: 0)) and (column_renames.__code__.co_argcount == 1):
            tds_row = LegendQLApiTdsRow.from_tds_frame("frame", self.__base_frame)
            try:
                result = column_renames(tds_row)
            except Exception as e:
                raise RuntimeError(
                    "rename' function column_renames argument lambda incompatible. "
                    "Error occurred while evaluating. Message: " + str(e)
                ) from e

            list_result = result if isinstance(result, list) else [result]
            for (i, r) in enumerate(list_result):
                if (isinstance(r, tuple) and (len(r) == 2) and
                        isinstance(r[0], LegendQLApiPrimitive) and isinstance(r[0].value(), PyLegendColumnExpression)
                        and isinstance(r[1], str)):
                    col_expr: PyLegendColumnExpression = r[0].value()
                    col_names.append(col_expr.get_column())
                    renamed_col_names.append(r[1])
                else:
                    raise TypeError(
                        "'rename' function column_renames argument lambda incompatible. Each element in rename list "
                        "should be a tuple with first element being a simple column expression and "
                        "second element being a string (renamed column name) "
                        f"(E.g - lambda r: [(r.c1, 'c2')]). Element at index {i} (0-indexed) is incompatible"
                    )

        else:
            raise TypeError("'rename' function column_renames argument can either be a list of renaming tuples or "
                            "a lambda function which takes one argument (LegendQLApiTdsRow)")

        self.__column_names = col_names
        self.__renamed_column_names = renamed_col_names

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        base_query = self.__base_frame.to_sql_query_object(config)

        db_extension = config.sql_to_string_generator().get_db_extension()
        quoted_column_names_to_change = [db_extension.quote_identifier(s) for s in self.__column_names]
        quoted_renamed_column_names = [db_extension.quote_identifier(s) for s in self.__renamed_column_names]

        new_select_items: PyLegendList[SelectItem] = []
        for col in base_query.select.selectItems:
            if not isinstance(col, SingleColumn):
                raise ValueError("Rename columns operation not supported for queries "
                                 "with columns other than SingleColumn")  # pragma: no cover
            if col.alias is None:
                raise ValueError("Rename columns operation not supported for queries "
                                 "with SingleColumns with missing alias")  # pragma: no cover
            if col.alias in quoted_column_names_to_change:
                new_alias = quoted_renamed_column_names[quoted_column_names_to_change.index(col.alias)]
                new_select_items.append(SingleColumn(alias=new_alias, expression=col.expression))
            else:
                new_select_items.append(col)

        new_query = copy_query(base_query)
        new_query.select.selectItems = new_select_items
        return new_query

    def to_pure(self, config: FrameToPureConfig) -> str:
        return (f"{self.__base_frame.to_pure(config)}{config.separator(1)}" +
                f"{config.separator(1)}".join([
                    f"->rename(~{x}, ~{y})"
                    for x, y in zip(
                        map(escape_column_name, self.__column_names),
                        map(escape_column_name, self.__renamed_column_names)
                    )
                ]))

    def base_frame(self) -> LegendQLApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["LegendQLApiBaseTdsFrame"]:
        return []

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        new_columns = []
        for base_col in self.__base_frame.columns():
            if base_col.get_name() in self.__column_names:
                renamed_column_name = self.__renamed_column_names[self.__column_names.index(base_col.get_name())]
                new_columns.append(base_col.copy_with_changed_name(renamed_column_name))
            else:
                new_columns.append(base_col.copy())
        return new_columns

    def validate(self) -> bool:
        if len(self.__column_names) != len(self.__renamed_column_names):
            raise ValueError(
                "column_names list and renamed_column_names list should have same size when renaming columns.\n"
                f"column_names list - (Count: {len(self.__column_names)}) - {self.__column_names}\n"
                f"renamed_column_names_list - (Count: {len(self.__renamed_column_names)}) - {self.__renamed_column_names}\n"
            )

        if len(self.__column_names) != len(set(self.__column_names)):
            raise ValueError(
                "column_names list shouldn't have duplicates when renaming columns.\n"
                f"column_names list - (Count: {len(self.__column_names)}) - {self.__column_names}\n"
            )

        if len(self.__renamed_column_names) != len(set(self.__renamed_column_names)):
            raise ValueError(
                "renamed_column_names_list list shouldn't have duplicates when renaming columns.\n"
                f"renamed_column_names_list - (Count: {len(self.__renamed_column_names)}) - {self.__renamed_column_names}\n"
            )

        return True



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
)
from pylegend.core.language.shared.expression import (
    PyLegendExpressionIntegerReturn,
    PyLegendExpressionStringReturn,
)
from pylegend.core.language.shared.operations.binary_expression import PyLegendBinaryExpression
from pylegend.core.language.shared.operations.unary_expression import PyLegendUnaryExpression
from pylegend.core.language.shared.helpers import generate_pure_functional_call
from pylegend.core.sql.metamodel import (
    Expression,
    QuerySpecification,
    ArithmeticType,
    ArithmeticExpression,
    NegativeExpression,
    FunctionCall,
    QualifiedName
)
from pylegend.core.sql.metamodel_extension import (
    AbsoluteExpression,
)
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.tds_frame import FrameToPureConfig


__all__: PyLegendSequence[str] = [
    "PyLegendIntegerAddExpression",
    "PyLegendIntegerAbsoluteExpression",
    "PyLegendIntegerNegativeExpression",
    "PyLegendIntegerSubtractExpression",
    "PyLegendIntegerMultiplyExpression",
    "PyLegendIntegerModuloExpression",
    "PyLegendIntegerCharExpression",
]


class PyLegendIntegerAddExpression(PyLegendBinaryExpression, PyLegendExpressionIntegerReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ArithmeticExpression(ArithmeticType.ADD, expression1, expression2)

    @staticmethod
    def __to_pure_func(op1_expr: str, op2_expr: str, config: FrameToPureConfig) -> str:
        return f"({op1_expr} + {op2_expr})"

    def __init__(self, operand1: PyLegendExpressionIntegerReturn, operand2: PyLegendExpressionIntegerReturn) -> None:
        PyLegendExpressionIntegerReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendIntegerAddExpression.__to_sql_func,
            PyLegendIntegerAddExpression.__to_pure_func,
            non_nullable=True,
            first_operand_needs_to_be_non_nullable=True,
            second_operand_needs_to_be_non_nullable=True
        )


class PyLegendIntegerSubtractExpression(PyLegendBinaryExpression, PyLegendExpressionIntegerReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ArithmeticExpression(ArithmeticType.SUBTRACT, expression1, expression2)

    @staticmethod
    def __to_pure_func(op1_expr: str, op2_expr: str, config: FrameToPureConfig) -> str:
        return f"({op1_expr} - {op2_expr})"

    def __init__(self, operand1: PyLegendExpressionIntegerReturn, operand2: PyLegendExpressionIntegerReturn) -> None:
        PyLegendExpressionIntegerReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendIntegerSubtractExpression.__to_sql_func,
            PyLegendIntegerSubtractExpression.__to_pure_func,
            non_nullable=True,
            first_operand_needs_to_be_non_nullable=True,
            second_operand_needs_to_be_non_nullable=True
        )


class PyLegendIntegerMultiplyExpression(PyLegendBinaryExpression, PyLegendExpressionIntegerReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ArithmeticExpression(ArithmeticType.MULTIPLY, expression1, expression2)

    @staticmethod
    def __to_pure_func(op1_expr: str, op2_expr: str, config: FrameToPureConfig) -> str:
        return f"({op1_expr} * {op2_expr})"

    def __init__(self, operand1: PyLegendExpressionIntegerReturn, operand2: PyLegendExpressionIntegerReturn) -> None:
        PyLegendExpressionIntegerReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendIntegerMultiplyExpression.__to_sql_func,
            PyLegendIntegerMultiplyExpression.__to_pure_func,
            non_nullable=True,
            first_operand_needs_to_be_non_nullable=True,
            second_operand_needs_to_be_non_nullable=True
        )


class PyLegendIntegerModuloExpression(PyLegendBinaryExpression, PyLegendExpressionIntegerReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ArithmeticExpression(
            ArithmeticType.MODULUS,
            ArithmeticExpression(
                ArithmeticType.ADD,
                ArithmeticExpression(ArithmeticType.MODULUS, expression1, expression2),
                expression2),
            expression2)

    @staticmethod
    def __to_pure_func(op1_expr: str, op2_expr: str, config: FrameToPureConfig) -> str:
        return generate_pure_functional_call("mod", [op1_expr, op2_expr])

    def __init__(self, operand1: PyLegendExpressionIntegerReturn, operand2: PyLegendExpressionIntegerReturn) -> None:
        PyLegendExpressionIntegerReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendIntegerModuloExpression.__to_sql_func,
            PyLegendIntegerModuloExpression.__to_pure_func,
            non_nullable=True,
            first_operand_needs_to_be_non_nullable=True,
            second_operand_needs_to_be_non_nullable=True
        )


class PyLegendIntegerAbsoluteExpression(PyLegendUnaryExpression, PyLegendExpressionIntegerReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return AbsoluteExpression(expression)

    @staticmethod
    def __to_pure_func(op_expr: str, config: FrameToPureConfig) -> str:
        return generate_pure_functional_call("abs", [op_expr])

    def __init__(self, operand: PyLegendExpressionIntegerReturn) -> None:
        PyLegendExpressionIntegerReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendIntegerAbsoluteExpression.__to_sql_func,
            PyLegendIntegerAbsoluteExpression.__to_pure_func,
            non_nullable=True,
            operand_needs_to_be_non_nullable=True,
        )


class PyLegendIntegerNegativeExpression(PyLegendUnaryExpression, PyLegendExpressionIntegerReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return NegativeExpression(expression)

    @staticmethod
    def __to_pure_func(op_expr: str, config: FrameToPureConfig) -> str:
        return generate_pure_functional_call("minus", [op_expr])

    def __init__(self, operand: PyLegendExpressionIntegerReturn) -> None:
        PyLegendExpressionIntegerReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendIntegerNegativeExpression.__to_sql_func,
            PyLegendIntegerNegativeExpression.__to_pure_func,
            non_nullable=True,
            operand_needs_to_be_non_nullable=True,
        )


class PyLegendIntegerCharExpression(PyLegendUnaryExpression, PyLegendExpressionStringReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return FunctionCall(
            name=QualifiedName(parts=["CHR"]),
            distinct=False,
            arguments=[expression],
            filter_=None, window=None
        )

    @staticmethod
    def __to_pure_func(op_expr: str, config: FrameToPureConfig) -> str:
        return generate_pure_functional_call("char", [op_expr])

    def __init__(self, operand: PyLegendExpressionIntegerReturn) -> None:
        PyLegendExpressionStringReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendIntegerCharExpression.__to_sql_func,
            PyLegendIntegerCharExpression.__to_pure_func,
            non_nullable=True,
            operand_needs_to_be_non_nullable=True,
        )



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
    PyLegendUnion,
)
from pylegend.core.language.shared.primitives.primitive import PyLegendPrimitive
from pylegend.core.language.shared.expression import PyLegendExpressionBooleanReturn
from pylegend.core.language.shared.literal_expressions import PyLegendBooleanLiteralExpression
from pylegend.core.language.shared.operations.boolean_operation_expressions import (
    PyLegendBooleanOrExpression,
    PyLegendBooleanAndExpression,
    PyLegendBooleanNotExpression,
)
from pylegend.core.sql.metamodel import (
    Expression,
    QuerySpecification
)
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.tds_frame import FrameToPureConfig


__all__: PyLegendSequence[str] = [
    "PyLegendBoolean"
]


class PyLegendBoolean(PyLegendPrimitive):
    __value: PyLegendExpressionBooleanReturn

    def __init__(
            self,
            value: PyLegendExpressionBooleanReturn
    ) -> None:
        self.__value = value

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return self.__value.to_sql_expression(frame_name_to_base_query_map, config)

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return self.__value.to_pure_expression(config)

    def value(self) -> PyLegendExpressionBooleanReturn:
        return self.__value

    def __or__(self, other: PyLegendUnion[bool, "PyLegendBoolean"]) -> "PyLegendBoolean":
        PyLegendBoolean.__validate__param_to_be_bool(other, "Boolean OR (|) parameter")
        other_op = PyLegendBooleanLiteralExpression(other) if isinstance(other, bool) else other.__value
        return PyLegendBoolean(PyLegendBooleanOrExpression(self.__value, other_op))

    def __ror__(self, other: PyLegendUnion[bool, "PyLegendBoolean"]) -> "PyLegendBoolean":
        PyLegendBoolean.__validate__param_to_be_bool(other, "Boolean OR (|) parameter")
        other_op = PyLegendBooleanLiteralExpression(other) if isinstance(other, bool) else other.__value
        return PyLegendBoolean(PyLegendBooleanOrExpression(other_op, self.__value))

    def __and__(self, other: PyLegendUnion[bool, "PyLegendBoolean"]) -> "PyLegendBoolean":
        PyLegendBoolean.__validate__param_to_be_bool(other, "Boolean AND (&) parameter")
        other_op = PyLegendBooleanLiteralExpression(other) if isinstance(other, bool) else other.__value
        return PyLegendBoolean(PyLegendBooleanAndExpression(self.__value, other_op))

    def __rand__(self, other: PyLegendUnion[bool, "PyLegendBoolean"]) -> "PyLegendBoolean":
        PyLegendBoolean.__validate__param_to_be_bool(other, "Boolean AND (&) parameter")
        other_op = PyLegendBooleanLiteralExpression(other) if isinstance(other, bool) else other.__value
        return PyLegendBoolean(PyLegendBooleanAndExpression(other_op, self.__value))

    def __invert__(self) -> "PyLegendBoolean":
        return PyLegendBoolean(PyLegendBooleanNotExpression(self.__value))

    @staticmethod
    def __validate__param_to_be_bool(param: PyLegendUnion[bool, "PyLegendBoolean"], desc: str) -> None:
        if not isinstance(param, (bool, PyLegendBoolean)):
            raise TypeError(desc + " should be a bool or a boolean expression (PyLegendBoolean)."
                                   " Got value " + str(param) + " of type: " + str(type(param)))



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import date, datetime
from pylegend._typing import (
    PyLegendList,
    PyLegendSequence,
    PyLegendTuple,
    PyLegendCallable,
    PyLegendUnion,
)
from pylegend.core.language import (
    PyLegendPrimitiveOrPythonPrimitive,
    PyLegendPrimitiveCollection,
    PyLegendPrimitive,
    create_primitive_collection,
    convert_literal_to_literal_expression
)
from pylegend.core.language.legendql_api.legendql_api_custom_expressions import LegendQLApiPrimitive
from pylegend.core.language.legendql_api.legendql_api_tds_row import LegendQLApiTdsRow
from pylegend.core.tds.abstract.function_helpers import tds_column_for_primitive
from pylegend.core.tds.legendql_api.frames.functions.legendql_api_function_helpers import infer_columns_from_frame
from pylegend.core.tds.legendql_api.frames.legendql_api_applied_function_tds_frame import LegendQLApiAppliedFunction
from pylegend.core.tds.sql_query_helpers import copy_query, create_sub_query
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    SingleColumn,
    SelectItem
)
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.tds_frame import FrameToPureConfig
from pylegend.core.tds.legendql_api.frames.legendql_api_base_tds_frame import LegendQLApiBaseTdsFrame
from pylegend.core.language.shared.helpers import escape_column_name, generate_pure_lambda

__all__: PyLegendSequence[str] = [
    "LegendQLApiGroupByFunction"
]


class LegendQLApiGroupByFunction(LegendQLApiAppliedFunction):
    __base_frame: LegendQLApiBaseTdsFrame
    __grouping_column_name_list: PyLegendList[str]
    __aggregates_list: PyLegendList[PyLegendTuple[str, PyLegendPrimitiveOrPythonPrimitive, PyLegendPrimitive]]

    @classmethod
    def name(cls) -> str:
        return "groupBy"

    def __init__(
            self,
            base_frame: LegendQLApiBaseTdsFrame,
            grouping_columns: PyLegendUnion[
                str,
                PyLegendList[str],
                PyLegendCallable[
                    [LegendQLApiTdsRow],
                    PyLegendUnion[LegendQLApiPrimitive, PyLegendList[LegendQLApiPrimitive]]
                ]
            ],
            aggregate_specifications: PyLegendUnion[
                PyLegendTuple[
                    str,
                    PyLegendCallable[[LegendQLApiTdsRow], PyLegendPrimitiveOrPythonPrimitive],
                    PyLegendCallable[[PyLegendPrimitiveCollection], PyLegendPrimitive]
                ],
                PyLegendList[
                    PyLegendTuple[
                        str,
                        PyLegendCallable[[LegendQLApiTdsRow], PyLegendPrimitiveOrPythonPrimitive],
                        PyLegendCallable[[PyLegendPrimitiveCollection], PyLegendPrimitive]
                    ]
                ]
            ]
    ) -> None:
        self.__base_frame = base_frame
        self.__grouping_column_name_list = infer_columns_from_frame(
            base_frame, grouping_columns, "'group_by' function grouping columns"
        )

        tds_row = LegendQLApiTdsRow.from_tds_frame("r", self.__base_frame)
        list_result = (
            aggregate_specifications if isinstance(aggregate_specifications, list) else [aggregate_specifications]
        )
        aggregates_list: PyLegendList[PyLegendTuple[str, PyLegendPrimitiveOrPythonPrimitive, PyLegendPrimitive]] = []
        for (i, agg_spec) in enumerate(list_result):
            error = (
                "'group_by' function aggregate specifications incompatible. "
                "Each aggregate specification should be a triplet with first element being the aggregation column "
                "name, second element being a mapper function (single argument lambda) and third element being the "
                "aggregation function (single argument lambda). "
                "E.g - ('count_col', lambda r: r['col1'], lambda c: c.count()). "
                f"Element at index {i} (0-indexed) is incompatible"
            )

            if isinstance(agg_spec, tuple) and isinstance(agg_spec[0], str):

                if not isinstance(agg_spec[1], type(lambda x: 0)) or (agg_spec[1].__code__.co_argcount != 1):
                    raise TypeError(error)

                try:
                    map_result = agg_spec[1](tds_row)
                except Exception as e:
                    raise RuntimeError(
                        "'group_by' function aggregate specifications incompatible. "
                        f"Error occurred while evaluating mapper lambda in the aggregate specification at "
                        f"index {i} (0-indexed). Message: " + str(e)
                    ) from e

                if not isinstance(map_result, (int, float, bool, str, date, datetime, PyLegendPrimitive)):
                    raise TypeError(
                        "'group_by' function aggregate specifications incompatible. "
                        f"Mapper lambda in the aggregate specification at index {i} (0-indexed) "
                        f"returns non-primitive - {str(type(map_result))}"
                    )

                collection = create_primitive_collection(map_result)

                if not isinstance(agg_spec[2], type(lambda x: 0)) or (agg_spec[2].__code__.co_argcount != 1):
                    raise TypeError(error)

                try:
                    agg_result = agg_spec[2](collection)
                except Exception as e:
                    raise RuntimeError(
                        "'group_by' function aggregate specifications incompatible. "
                        f"Error occurred while evaluating aggregation lambda in the aggregate specification at "
                        f"index {i} (0-indexed). Message: " + str(e)
                    ) from e

                if not isinstance(agg_result, PyLegendPrimitive):
                    raise TypeError(
                        "'group_by' function aggregate specifications incompatible. "
                        f"Aggregation lambda in the aggregate specification at index {i} (0-indexed) "
                        f"returns non-primitive - {str(type(agg_result))}"
                    )

                aggregates_list.append((agg_spec[0], map_result, agg_result))

            else:
                raise TypeError(error)

        self.__aggregates_list = aggregates_list

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        db_extension = config.sql_to_string_generator().get_db_extension()
        base_query = self.__base_frame.to_sql_query_object(config)

        should_create_sub_query = (len(base_query.groupBy) > 0) or base_query.select.distinct or \
                                  (base_query.offset is not None) or (base_query.limit is not None)

        columns_to_retain = [db_extension.quote_identifier(x) for x in self.__grouping_column_name_list]
        if should_create_sub_query:
            new_query = create_sub_query(base_query, config, "root")
        else:
            new_query = copy_query(base_query)

        new_cols_with_index: PyLegendList[PyLegendTuple[int, 'SelectItem']] = []
        for col in new_query.select.selectItems:
            if not isinstance(col, SingleColumn):
                raise ValueError("Group By operation not supported for queries "
                                 "with columns other than SingleColumn")  # pragma: no cover
            if col.alias is None:
                raise ValueError("Group By operation not supported for queries "
                                 "with SingleColumns with missing alias")  # pragma: no cover
            if col.alias in columns_to_retain:
                new_cols_with_index.append((columns_to_retain.index(col.alias), col))

        new_select_items = [y[1] for y in sorted(new_cols_with_index, key=lambda x: x[0])]

        tds_row = LegendQLApiTdsRow.from_tds_frame("r", self.__base_frame)
        for agg in self.__aggregates_list:
            agg_sql_expr = agg[2].to_sql_expression({"r": new_query}, config)

            new_select_items.append(
                SingleColumn(alias=db_extension.quote_identifier(agg[0]), expression=agg_sql_expr)
            )

        new_query.select.selectItems = new_select_items
        new_query.groupBy = [
            (lambda x: x[c])(tds_row).to_sql_expression({"r": new_query}, config)
            for c in self.__grouping_column_name_list
        ]
        return new_query

    def to_pure(self, config: FrameToPureConfig) -> str:
        group_strings = []
        for col_name in self.__grouping_column_name_list:
            group_strings.append(escape_column_name(col_name))

        agg_strings = []
        for agg in self.__aggregates_list:
            map_expr_string = (agg[1].to_pure_expression(config) if isinstance(agg[1], PyLegendPrimitive)
                               else convert_literal_to_literal_expression(agg[1]).to_pure_expression(config))
            agg_expr_string = agg[2].to_pure_expression(config).replace(map_expr_string, "$c")
            agg_strings.append(f"{escape_column_name(agg[0])}:{generate_pure_lambda('r', map_expr_string)}:"
                               f"{generate_pure_lambda('c', agg_expr_string)}")

        return (f"{self.__base_frame.to_pure(config)}{config.separator(1)}" +
                f"->groupBy({config.separator(2)}"
                f"~[{', '.join(group_strings)}],{config.separator(2, True)}"
                f"~[{', '.join(agg_strings)}]{config.separator(1)}"
                f")")

    def base_frame(self) -> LegendQLApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["LegendQLApiBaseTdsFrame"]:
        return []

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        base_columns = self.__base_frame.columns()
        new_columns = []
        for c in self.__grouping_column_name_list:
            for base_col in base_columns:
                if base_col.get_name() == c:
                    new_columns.append(base_col.copy())
                    break
        for agg in self.__aggregates_list:
            new_columns.append(tds_column_for_primitive(agg[0], agg[2]))
        return new_columns

    def validate(self) -> bool:
        base_columns = self.__base_frame.columns()
        for c in self.__grouping_column_name_list:
            found_col = False
            for base_col in base_columns:
                if base_col.get_name() == c:
                    found_col = True
                    break
            if not found_col:
                raise ValueError(
                    f"Column - '{c}' in group_by columns list doesn't exist in the current frame. "
                    f"Current frame columns: {[x.get_name() for x in base_columns]}"
                )

        agg_cols = [c[0] for c in self.__aggregates_list]
        new_cols = self.__grouping_column_name_list + agg_cols
        if len(new_cols) == 0:
            raise ValueError("At-least one grouping column or aggregate specification must be provided "
                             "when using group_by function")
        if len(new_cols) != len(set(new_cols)):
            raise ValueError("Found duplicate column names in grouping columns and aggregation columns. "
                             f"Grouping columns - {self.__grouping_column_name_list}, Aggregation columns - {agg_cols}")
        return True



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABCMeta
from pylegend._typing import (
    PyLegendSequence,
    PyLegendList,
)
from io import StringIO
from pylegend.core.tds.tds_column import (
    PrimitiveType,
    PrimitiveTdsColumn)
from pylegend.core.tds.tds_frame import FrameToPureConfig, FrameToSqlConfig, PyLegendTdsFrame
from pylegend.core.sql.metamodel import (
    QuerySpecification,
)
import pandas as pd

__all__: PyLegendSequence[str] = [
    "CsvInputFrameAbstract",
    "tds_columns_from_csv_string"
]


class CsvInputFrameAbstract(PyLegendTdsFrame, metaclass=ABCMeta):
    __csv_string: str

    def __init__(
            self,
            csv_string: str,
    ) -> None:
        super().__init__(columns=tds_columns_from_csv_string(csv_string))  # type: ignore[call-arg]
        self.__csv_string = csv_string

    def to_sql_query_object(self, config: FrameToSqlConfig) -> QuerySpecification:
        raise RuntimeError("SQL generation for csv tds frames is not supported yet.")

    def to_pure(self, config: FrameToPureConfig) -> str:
        return f"#TDS\n{self.__csv_string}#"


def tds_columns_from_csv_string(
        csv_string: str
) -> PyLegendList[PrimitiveTdsColumn]:
    df = pd.read_csv(StringIO(csv_string))
    tds_columns = []
    dt = pd.api.types

    for col in df.columns:
        dtype = df[col].dtype

        if dt.is_bool_dtype(dtype):
            primitive_type = PrimitiveType.Boolean

        elif dt.is_integer_dtype(dtype):
            primitive_type = PrimitiveType.Integer

        elif dt.is_float_dtype(dtype):
            primitive_type = PrimitiveType.Float

        elif is_strict_date_or_datetime(df[col]):
            primitive_type = PrimitiveType.Date

        else:
            primitive_type = PrimitiveType.String

        tds_columns.append(
            PrimitiveTdsColumn(name=col, _type=primitive_type)
        )

    return tds_columns


def is_strict_date_or_datetime(col: pd.Series) -> bool:  # type: ignore[explicit-any]
    try:
        pd.to_datetime(col, format="%Y-%m-%d %H:%M:%S", exact=True, errors="raise")
        return True
    except (ValueError, TypeError):
        pass

    try:
        pd.to_datetime(col, format="%Y-%m-%d", exact=True, errors="raise")
        return True
    except (ValueError, TypeError):
        return False



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta
from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
    PyLegendCallable,
)
from pylegend.core.language.shared.expression import (
    PyLegendExpression,
)
from pylegend.core.language.shared.helpers import expr_has_matching_start_and_end_parentheses
from pylegend.core.sql.metamodel import (
    Expression,
    QuerySpecification,
)
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.tds_frame import FrameToPureConfig


__all__: PyLegendSequence[str] = [
    "PyLegendBinaryExpression"
]


class PyLegendBinaryExpression(PyLegendExpression, metaclass=ABCMeta):
    __operand1: PyLegendExpression
    __operand2: PyLegendExpression
    __to_sql_func: PyLegendCallable[
        [Expression, Expression, PyLegendDict[str, QuerySpecification], FrameToSqlConfig],
        Expression
    ]
    __to_pure_func: PyLegendCallable[[str, str, FrameToPureConfig], str]
    __non_nullable: bool
    __first_operand_needs_to_be_non_nullable: bool
    __second_operand_needs_to_be_non_nullable: bool

    def __init__(
            self,
            operand1: PyLegendExpression,
            operand2: PyLegendExpression,
            to_sql_func: PyLegendCallable[
                [Expression, Expression, PyLegendDict[str, QuerySpecification], FrameToSqlConfig],
                Expression
            ],
            to_pure_func: PyLegendCallable[[str, str, FrameToPureConfig], str],
            non_nullable: bool = False,
            first_operand_needs_to_be_non_nullable: bool = False,
            second_operand_needs_to_be_non_nullable: bool = False,
    ) -> None:
        self.__operand1 = operand1
        self.__operand2 = operand2
        self.__to_sql_func = to_sql_func
        self.__to_pure_func = to_pure_func
        self.__non_nullable = non_nullable
        self.__first_operand_needs_to_be_non_nullable = first_operand_needs_to_be_non_nullable
        self.__second_operand_needs_to_be_non_nullable = second_operand_needs_to_be_non_nullable

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        op1_expr = self.__operand1.to_sql_expression(frame_name_to_base_query_map, config)
        op2_expr = self.__operand2.to_sql_expression(frame_name_to_base_query_map, config)
        return self.__to_sql_func(
            op1_expr,
            op2_expr,
            frame_name_to_base_query_map,
            config
        )

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        op1_expr = self.__operand1.to_pure_expression(config)
        if self.__first_operand_needs_to_be_non_nullable:
            op1_expr = (
                op1_expr if self.__operand1.is_non_nullable() else
                f"toOne({op1_expr[1:-1] if expr_has_matching_start_and_end_parentheses(op1_expr) else op1_expr})"
            )
        op2_expr = self.__operand2.to_pure_expression(config)
        if self.__second_operand_needs_to_be_non_nullable:
            op2_expr = (
                op2_expr if self.__operand2.is_non_nullable() else
                f"toOne({op2_expr[1:-1] if expr_has_matching_start_and_end_parentheses(op2_expr) else op2_expr})"
            )
        return self.__to_pure_func(op1_expr, op2_expr, config)

    def is_non_nullable(self) -> bool:
        return self.__non_nullable



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta

from pylegend._typing import (
    PyLegendSequence,
    PyLegendTypeVar,
    PyLegendList,
    PyLegendOptional,
    PyLegendCallable,
    PyLegendUnion,
)
from pylegend.core.language import (
    LegacyApiTdsRow,
    PyLegendBoolean,
    PyLegendPrimitiveOrPythonPrimitive,
    LegacyApiAggregateSpecification,
)
from pylegend.core.tds.abstract.frames.base_tds_frame import BaseTdsFrame
from pylegend.core.tds.legacy_api.frames.legacy_api_tds_frame import LegacyApiTdsFrame
from pylegend.core.tds.tds_column import TdsColumn

__all__: PyLegendSequence[str] = [
    "LegacyApiBaseTdsFrame"
]

R = PyLegendTypeVar('R')


class LegacyApiBaseTdsFrame(LegacyApiTdsFrame, BaseTdsFrame, metaclass=ABCMeta):

    def __init__(self, columns: PyLegendSequence[TdsColumn]) -> None:
        BaseTdsFrame.__init__(self, columns=columns)

    def head(self, row_count: int = 5) -> "LegacyApiTdsFrame":
        from pylegend.core.tds.legacy_api.frames.legacy_api_applied_function_tds_frame import (
            LegacyApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legacy_api.frames.functions.legacy_api_head_function import (
            LegacyApiHeadFunction
        )
        return LegacyApiAppliedFunctionTdsFrame(LegacyApiHeadFunction(self, row_count))

    def take(self, row_count: int = 5) -> "LegacyApiTdsFrame":
        return self.head(row_count=row_count)

    def limit(self, row_count: int = 5) -> "LegacyApiTdsFrame":
        return self.head(row_count=row_count)

    def drop(self, row_count: int = 5) -> "LegacyApiTdsFrame":
        from pylegend.core.tds.legacy_api.frames.legacy_api_applied_function_tds_frame import (
            LegacyApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legacy_api.frames.functions.legacy_api_drop_function import (
            LegacyApiDropFunction
        )
        return LegacyApiAppliedFunctionTdsFrame(LegacyApiDropFunction(self, row_count))

    def slice(self, start_row: int, end_row_exclusive: int) -> "LegacyApiTdsFrame":
        from pylegend.core.tds.legacy_api.frames.legacy_api_applied_function_tds_frame import (
            LegacyApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legacy_api.frames.functions.legacy_api_slice_function import (
            LegacyApiSliceFunction
        )
        return LegacyApiAppliedFunctionTdsFrame(LegacyApiSliceFunction(self, start_row, end_row_exclusive))

    def distinct(self) -> "LegacyApiTdsFrame":
        from pylegend.core.tds.legacy_api.frames.legacy_api_applied_function_tds_frame import (
            LegacyApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legacy_api.frames.functions.legacy_api_distinct_function import (
            LegacyApiDistinctFunction
        )
        return LegacyApiAppliedFunctionTdsFrame(LegacyApiDistinctFunction(self))

    def restrict(self, column_name_list: PyLegendList[str]) -> "LegacyApiTdsFrame":
        from pylegend.core.tds.legacy_api.frames.legacy_api_applied_function_tds_frame import (
            LegacyApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legacy_api.frames.functions.legacy_api_restrict_function import (
            LegacyApiRestrictFunction
        )
        return LegacyApiAppliedFunctionTdsFrame(LegacyApiRestrictFunction(self, column_name_list))

    def sort(
            self,
            column_name_list: PyLegendList[str],
            direction_list: PyLegendOptional[PyLegendList[str]] = None
    ) -> "LegacyApiTdsFrame":
        from pylegend.core.tds.legacy_api.frames.legacy_api_applied_function_tds_frame import (
            LegacyApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legacy_api.frames.functions.legacy_api_sort_function import (
            LegacyApiSortFunction
        )
        return LegacyApiAppliedFunctionTdsFrame(LegacyApiSortFunction(self, column_name_list, direction_list))

    def concatenate(self, other: "LegacyApiTdsFrame") -> "LegacyApiTdsFrame":
        from pylegend.core.tds.legacy_api.frames.legacy_api_applied_function_tds_frame import (
            LegacyApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legacy_api.frames.functions.legacy_api_concatenate_function import (
            LegacyApiConcatenateFunction
        )
        return LegacyApiAppliedFunctionTdsFrame(LegacyApiConcatenateFunction(self, other))

    def rename_columns(
            self,
            column_names: PyLegendList[str],
            renamed_column_names: PyLegendList[str]
    ) -> "LegacyApiTdsFrame":
        from pylegend.core.tds.legacy_api.frames.legacy_api_applied_function_tds_frame import (
            LegacyApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legacy_api.frames.functions.legacy_api_rename_columns_function import (
            LegacyApiRenameColumnsFunction
        )
        return LegacyApiAppliedFunctionTdsFrame(LegacyApiRenameColumnsFunction(self, column_names, renamed_column_names))

    def filter(
            self,
            filter_function: PyLegendCallable[[LegacyApiTdsRow], PyLegendUnion[bool, PyLegendBoolean]]
    ) -> "LegacyApiTdsFrame":
        from pylegend.core.tds.legacy_api.frames.legacy_api_applied_function_tds_frame import (
            LegacyApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legacy_api.frames.functions.legacy_api_filter_function import (
            LegacyApiFilterFunction
        )
        return LegacyApiAppliedFunctionTdsFrame(LegacyApiFilterFunction(self, filter_function))

    def extend(
            self,
            functions_list: PyLegendList[PyLegendCallable[[LegacyApiTdsRow], PyLegendPrimitiveOrPythonPrimitive]],
            column_names_list: PyLegendList[str]
    ) -> "LegacyApiTdsFrame":
        from pylegend.core.tds.legacy_api.frames.legacy_api_applied_function_tds_frame import (
            LegacyApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legacy_api.frames.functions.legacy_api_extend_function import (
            LegacyApiExtendFunction
        )
        return LegacyApiAppliedFunctionTdsFrame(LegacyApiExtendFunction(self, functions_list, column_names_list))

    def join(
            self,
            other: "LegacyApiTdsFrame",
            join_condition: PyLegendCallable[[LegacyApiTdsRow, LegacyApiTdsRow], PyLegendUnion[bool, PyLegendBoolean]],
            join_type: str = 'LEFT_OUTER'
    ) -> "LegacyApiTdsFrame":
        from pylegend.core.tds.legacy_api.frames.legacy_api_applied_function_tds_frame import (
            LegacyApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legacy_api.frames.functions.legacy_api_join_function import (
            LegacyApiJoinFunction
        )
        return LegacyApiAppliedFunctionTdsFrame(
            LegacyApiJoinFunction(self, other, join_condition, join_type)
        )

    def join_by_columns(
            self,
            other: "LegacyApiTdsFrame",
            self_columns: PyLegendList[str],
            other_columns: PyLegendList[str],
            join_type: str = 'LEFT_OUTER'
    ) -> "LegacyApiTdsFrame":
        from pylegend.core.tds.legacy_api.frames.legacy_api_applied_function_tds_frame import (
            LegacyApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legacy_api.frames.functions.legacy_api_join_by_columns_function import (
            LegacyApiJoinByColumnsFunction
        )
        return LegacyApiAppliedFunctionTdsFrame(
            LegacyApiJoinByColumnsFunction(self, other, self_columns, other_columns, join_type)
        )

    def group_by(
            self,
            grouping_columns: PyLegendList[str],
            aggregations: PyLegendList[LegacyApiAggregateSpecification],
    ) -> "LegacyApiTdsFrame":
        from pylegend.core.tds.legacy_api.frames.legacy_api_applied_function_tds_frame import (
            LegacyApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legacy_api.frames.functions.legacy_api_group_by_function import (
            LegacyApiGroupByFunction
        )
        return LegacyApiAppliedFunctionTdsFrame(
            LegacyApiGroupByFunction(self, grouping_columns, aggregations)
        )



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendList,
    PyLegendSequence,
    PyLegendCallable,
    PyLegendUnion,
)
from pylegend.core.language import PyLegendColumnExpression
from pylegend.core.language.legendql_api.legendql_api_custom_expressions import (
    LegendQLApiPrimitive,
    LegendQLApiSortInfo,
    LegendQLApiSortDirection,
)
from pylegend.core.language.legendql_api.legendql_api_tds_row import LegendQLApiTdsRow
from pylegend.core.tds.legendql_api.frames.legendql_api_base_tds_frame import LegendQLApiBaseTdsFrame


__all__: PyLegendSequence[str] = [
    "infer_columns_from_frame",
    "infer_sorts_from_frame",
]


def infer_columns_from_frame(
    base_frame: LegendQLApiBaseTdsFrame,
    columns: PyLegendUnion[
        str,
        PyLegendList[str],
        PyLegendCallable[
            [LegendQLApiTdsRow],
            PyLegendUnion[LegendQLApiPrimitive, PyLegendList[LegendQLApiPrimitive]]
        ]
    ],
    context: str
) -> PyLegendList[str]:
    if isinstance(columns, str):
        return [columns]
    if isinstance(columns, list) and all([isinstance(c, str) for c in columns]):
        return columns

    if isinstance(columns, type(lambda x: 0)) and (columns.__code__.co_argcount == 1):
        tds_row = LegendQLApiTdsRow.from_tds_frame("frame", base_frame)
        try:
            result = columns(tds_row)
        except Exception as e:
            raise RuntimeError(
                f"{context} argument lambda incompatible. Error occurred while evaluating. Message: " + str(e)
            ) from e

        list_result = result if isinstance(result, list) else [result]

        columns_list = []
        for (i, r) in enumerate(list_result):
            if isinstance(r, LegendQLApiPrimitive) and isinstance(r.value(), PyLegendColumnExpression):
                col_expr: PyLegendColumnExpression = r.value()
                columns_list.append(col_expr.get_column())
            else:
                raise TypeError(
                    f"{context} argument lambda incompatible. Columns can be simple column expressions "
                    f"(E.g - lambda r: [r.column1, r.column2, r['column with spaces']). "
                    f"Element at index {i} (0-indexed) is incompatible"
                )
        return columns_list

    raise TypeError(f"{context} argument can either be a list of strings (column names) or "
                    f"a lambda function which takes one argument (LegendQLApiTdsRow)")


def infer_sorts_from_frame(
    base_frame: LegendQLApiBaseTdsFrame,
    sort_infos: PyLegendUnion[
        str,
        PyLegendList[str],
        PyLegendCallable[
            [LegendQLApiTdsRow],
            PyLegendUnion[
                LegendQLApiPrimitive,
                LegendQLApiSortInfo,
                PyLegendList[PyLegendUnion[LegendQLApiPrimitive, LegendQLApiSortInfo]],
            ]
        ]
    ],
    context: str
) -> PyLegendList[LegendQLApiSortInfo]:
    tds_row = LegendQLApiTdsRow.from_tds_frame("frame", base_frame)
    sort_info_list = []

    if isinstance(sort_infos, str):
        col_expr1: PyLegendColumnExpression = tds_row[sort_infos].value()  # type: ignore
        sort_info_list.append(LegendQLApiSortInfo(col_expr1, LegendQLApiSortDirection.ASC))

    elif isinstance(sort_infos, list) and all([isinstance(s, str) for s in sort_infos]):
        for s in sort_infos:
            col_expr2: PyLegendColumnExpression = tds_row[s].value()  # type: ignore
            sort_info_list.append(LegendQLApiSortInfo(col_expr2, LegendQLApiSortDirection.ASC))

    elif isinstance(sort_infos, type(lambda x: 0)) and (sort_infos.__code__.co_argcount == 1):
        try:
            result = sort_infos(tds_row)
        except Exception as e:
            raise RuntimeError(
                f"{context} argument lambda incompatible. "
                "Error occurred while evaluating. Message: " + str(e)
            ) from e

        list_result = result if isinstance(result, list) else [result]
        for (i, r) in enumerate(list_result):
            if isinstance(r, LegendQLApiPrimitive) and isinstance(r.value(), PyLegendColumnExpression):
                col_expr3: PyLegendColumnExpression = r.value()
                sort_info_list.append(LegendQLApiSortInfo(col_expr3, LegendQLApiSortDirection.ASC))
            elif isinstance(r, LegendQLApiSortInfo):
                sort_info_list.append(r)
            else:
                raise TypeError(
                    f"{context} argument lambda incompatible. Columns can be simple column "
                    f"expressions or sort infos. "
                    f"(E.g - lambda r: [r.column1, r['column with spaces'].descending(), r.column3.ascending()). "
                    f"Element at index {i} (0-indexed) is incompatible"
                )
    else:
        raise TypeError(f"{context} argument can either be a list of strings (column names) or "
                        "a lambda function which takes one argument (LegendQLApiTdsRow)")

    return sort_info_list



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pylegend._typing import (
    PyLegendIterator,
    PyLegendSequence,
    PyLegendOptional,
)

__all__: PyLegendSequence["str"] = [
    "ResponseReader"
]


class ResponseReader:
    __iter_content: PyLegendIterator[bytes]
    __byte_array: bytearray

    def __init__(self, iter_content: PyLegendIterator[bytes]):
        self.__iter_content = iter_content
        self.__byte_array = bytearray()

    def __iter__(self) -> PyLegendIterator[bytes]:
        return self.__iter_content

    def read(self, n: PyLegendOptional[int] = None) -> bytes:
        if n == 0:
            return b""
        elif n is None:
            # Should read everything
            self.__byte_array.extend(b"".join(self.__iter_content))
            return bytes(self.__byte_array)
        else:
            while len(self.__byte_array) < n:
                try:
                    self.__byte_array.extend(next(self.__iter_content))
                except StopIteration:
                    break
            read_bytes, self.__byte_array = bytes(self.__byte_array[:n]), self.__byte_array[n:]
            return read_bytes



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
from pylegend._typing import (
    PyLegendSequence,
)
from pylegend.core.tds.abstract.frames.applied_function_tds_frame import AppliedFunction, AppliedFunctionTdsFrame
from pylegend.core.tds.legendql_api.frames.legendql_api_base_tds_frame import LegendQLApiBaseTdsFrame


__all__: PyLegendSequence[str] = [
    "LegendQLApiAppliedFunctionTdsFrame",
    "LegendQLApiAppliedFunction",
]


class LegendQLApiAppliedFunction(AppliedFunction, metaclass=ABCMeta):
    @abstractmethod
    def tds_frame_parameters(self) -> PyLegendSequence["LegendQLApiBaseTdsFrame"]:
        pass  # pragma: no cover


class LegendQLApiAppliedFunctionTdsFrame(LegendQLApiBaseTdsFrame, AppliedFunctionTdsFrame):
    def __init__(self, applied_function: LegendQLApiAppliedFunction):
        AppliedFunctionTdsFrame.__init__(self, applied_function)



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend.core.request.service_client import (
    ServiceClient,
    RequestMethod
)
from pylegend.core.request.response_reader import ResponseReader
from pylegend.core.request.auth import AuthScheme, LocalhostEmptyAuthScheme
from pylegend._typing import (
    PyLegendSequence,
    PyLegendOptional,
)
from pylegend.core.tds.tds_column import TdsColumn, tds_columns_from_json


__all__: PyLegendSequence[str] = [
    "LegendClient",
]


class LegendClient(ServiceClient):
    def __init__(
            self,
            host: str,
            port: int,
            secure_http: bool = True,
            path_prefix: PyLegendOptional[str] = "/api",
            auth_scheme: AuthScheme = LocalhostEmptyAuthScheme(),
            retry_count: int = 2
    ) -> None:
        super().__init__(
            host=host,
            port=port,
            secure_http=secure_http,
            path_prefix=path_prefix,
            auth_scheme=auth_scheme,
            retry_count=retry_count
        )

    def get_sql_string_schema(
            self,
            sql: str
    ) -> PyLegendSequence[TdsColumn]:
        response = super()._execute_service(
            method=RequestMethod.POST,
            path="sql/v1/execution/getSchemaFromQueryString",
            data=sql,
            headers={"Content-Type": "text/plain"},
            stream=False
        )
        response_text: str = response.text
        return tds_columns_from_json(response_text)

    def execute_sql_string(
            self,
            sql: str,
            chunk_size: PyLegendOptional[int] = None
    ) -> ResponseReader:
        iter_content = super()._execute_service(
            method=RequestMethod.POST,
            path="sql/v1/execution/executeQueryString",
            data=sql,
            headers={"Content-Type": "text/plain"},
            stream=True
        ).iter_content(chunk_size=chunk_size)
        return ResponseReader(iter_content)

    def parse_model(
            self,
            model_code: str,
            return_source_information: bool = False
    ) -> str:
        response = super()._execute_service(
            method=RequestMethod.POST,
            path="pure/v1/grammar/grammarToJson/model",
            data=model_code,
            headers={"Content-Type": "text/plain"},
            query_params=[("returnSourceInformation", "true" if return_source_information else "false")]
        )
        return response.text

    def compile_model(
            self,
            model_json: str
    ) -> None:
        super()._execute_service(
            method=RequestMethod.POST,
            path="pure/v1/compilation/compile",
            data=model_json,
            headers={"Content-Type": "application/json"}
        )

    def parse_and_compile_model(
            self,
            model_code: str
    ) -> None:
        parse_response = self.parse_model(model_code)
        self.compile_model(parse_response)

    def __eq__(self, other: 'object') -> bool:
        if self is other:
            return True
        if isinstance(other, LegendClient):
            return self.get_host() == other.get_host() and self.get_port() == other.get_port()
        return False



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendList,
    PyLegendSequence,
)
from pylegend.core.tds.legendql_api.frames.legendql_api_applied_function_tds_frame import LegendQLApiAppliedFunction
from pylegend.core.tds.sql_query_helpers import copy_query, create_sub_query
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    LongLiteral,
)
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.tds_frame import FrameToPureConfig
from pylegend.core.tds.legendql_api.frames.legendql_api_base_tds_frame import LegendQLApiBaseTdsFrame


__all__: PyLegendSequence[str] = [
    "LegendQLApiHeadFunction"
]


class LegendQLApiHeadFunction(LegendQLApiAppliedFunction):
    __base_frame: LegendQLApiBaseTdsFrame
    __row_count: int

    @classmethod
    def name(cls) -> str:
        return "head"

    def __init__(self, base_frame: LegendQLApiBaseTdsFrame, row_count: int) -> None:
        self.__base_frame = base_frame
        self.__row_count = row_count

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        base_query = self.__base_frame.to_sql_query_object(config)
        should_create_sub_query = (base_query.limit is not None)
        new_query = (
            create_sub_query(base_query, config, "root") if should_create_sub_query else
            copy_query(base_query)
        )
        new_query.limit = LongLiteral(value=self.__row_count)
        return new_query

    def to_pure(self, config: FrameToPureConfig) -> str:
        return (f"{self.__base_frame.to_pure(config)}{config.separator(1)}"
                f"->limit({self.__row_count})")

    def base_frame(self) -> LegendQLApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["LegendQLApiBaseTdsFrame"]:
        return []

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        return [c.copy() for c in self.__base_frame.columns()]

    def validate(self) -> bool:
        if self.__row_count < 0:
            raise ValueError("Row count argument of head/limit function cannot be negative")
        return True



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pylegend._typing import (
    PyLegendSequence,
)
from pylegend.core.request import LegendClient
from pylegend.core.project_cooridnates import ProjectCoordinates
from pylegend.core.tds.legendql_api.frames.legendql_api_tds_frame import LegendQLApiTdsFrame


__all__: PyLegendSequence[str] = [
    "LegendQLApiTdsClient",
    "legendql_api_tds_client",
]


class LegendQLApiTdsClient:
    __legend_client: LegendClient

    def __init__(
            self,
            legend_client: LegendClient
    ) -> None:
        self.__legend_client = legend_client

    def legend_service_frame(
            self,
            service_pattern: str,
            project_coordinates: ProjectCoordinates
    ) -> LegendQLApiTdsFrame:
        from pylegend.extensions.tds.legendql_api.frames.legendql_api_legend_service_input_frame import (
            LegendQLApiLegendServiceInputFrame
        )
        return LegendQLApiLegendServiceInputFrame(
            pattern=service_pattern,
            project_coordinates=project_coordinates,
            legend_client=self.__legend_client
        )

    def legend_function_frame(
            self,
            function_path: str,
            project_coordinates: ProjectCoordinates
    ) -> LegendQLApiTdsFrame:
        from pylegend.extensions.tds.legendql_api.frames.legendql_api_legend_function_input_frame import (
            LegendQLApiLegendFunctionInputFrame
        )
        return LegendQLApiLegendFunctionInputFrame(
            path=function_path,
            project_coordinates=project_coordinates,
            legend_client=self.__legend_client
        )


def legendql_api_tds_client(
        legend_client: LegendClient
) -> LegendQLApiTdsClient:
    return LegendQLApiTdsClient(
        legend_client=legend_client
    )



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime, date
from pylegend._typing import (
    PyLegendSequence,
    PyLegendUnion,
)
from pylegend.core.language import (
    convert_literal_to_literal_expression,
    PyLegendBooleanLiteralExpression,
    PyLegendStringLiteralExpression,
    PyLegendIntegerLiteralExpression,
    PyLegendFloatLiteralExpression,
    PyLegendStrictDateLiteralExpression,
    PyLegendDateTimeLiteralExpression,
    PyLegendFloat,
    PyLegendDateTime,
    PyLegendStrictDate,
    PyLegendString,
    PyLegendInteger,
    PyLegendBoolean,
    PyLegendPrimitive
)

__all__: PyLegendSequence[str] = [
    "c",
]


def c(constant: PyLegendUnion[int, float, bool, str, datetime, date]) -> PyLegendPrimitive:
    expr = convert_literal_to_literal_expression(constant)
    if isinstance(expr, PyLegendBooleanLiteralExpression):
        return PyLegendBoolean(expr)
    if isinstance(expr, PyLegendStringLiteralExpression):
        return PyLegendString(expr)
    if isinstance(expr, PyLegendIntegerLiteralExpression):
        return PyLegendInteger(expr)
    if isinstance(expr, PyLegendFloatLiteralExpression):
        return PyLegendFloat(expr)
    if isinstance(expr, PyLegendDateTimeLiteralExpression):
        return PyLegendDateTime(expr)
    if isinstance(expr, PyLegendStrictDateLiteralExpression):
        return PyLegendStrictDate(expr)
    raise RuntimeError(f"Unsupported constant type: {type(constant)}")



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod

from requests import PreparedRequest
from requests.cookies import create_cookie

from pylegend._typing import (
    PyLegendSequence,
    PyLegendOptional,
    PyLegendCallable,
    PyLegendDict,
)
from requests.auth import AuthBase

__all__: PyLegendSequence[str] = [
    "AuthScheme",
    "LocalhostEmptyAuthScheme",
    "HeaderTokenAuthScheme",
    "CookieAuthScheme",
]


class AuthScheme(metaclass=ABCMeta):

    @abstractmethod
    def get_auth_base(self) -> PyLegendOptional[AuthBase]:
        pass


class LocalhostEmptyAuthScheme(AuthScheme):

    def get_auth_base(self) -> PyLegendOptional[AuthBase]:
        return None


class HeaderTokenAuth(AuthBase):
    __header_name: str
    __token_provider: PyLegendCallable[[], str]
    __query_params: PyLegendOptional[PyLegendDict[str, str]]

    def __init__(
            self,
            header_name: str,
            token_provider: PyLegendCallable[[], str],
            query_params: PyLegendOptional[PyLegendDict[str, str]] = None
    ) -> None:
        self.__header_name = header_name
        self.__token_provider = token_provider
        self.__query_params = query_params

    def __call__(self, r: PreparedRequest) -> PreparedRequest:
        if self.__query_params:
            r.prepare_url(r.url, self.__query_params)
        r.headers[self.__header_name] = self.__token_provider()
        return r


class HeaderTokenAuthScheme(AuthScheme):
    __header_name: str
    __token_provider: PyLegendCallable[[], str]
    __query_params: PyLegendOptional[PyLegendDict[str, str]]

    def __init__(
            self,
            header_name: str,
            token_provider: PyLegendCallable[[], str],
            query_params: PyLegendOptional[PyLegendDict[str, str]] = None
    ) -> None:
        self.__header_name = header_name
        self.__token_provider = token_provider
        self.__query_params = query_params

    def get_auth_base(self) -> PyLegendOptional[AuthBase]:
        return HeaderTokenAuth(self.__header_name, self.__token_provider, self.__query_params)


class CookieAuth(AuthBase):
    __cookie_name: str
    __cookie_provider: PyLegendCallable[[], str]
    __query_params: PyLegendOptional[PyLegendDict[str, str]]

    def __init__(  # type: ignore
            self,
            cookie_name: str,
            cookie_provider: PyLegendCallable[[], str],
            query_params: PyLegendOptional[PyLegendDict[str, str]] = None,
            **extra_cookie_params
    ) -> None:
        self.__cookie_name = cookie_name
        self.__cookie_provider = cookie_provider
        self.__query_params = query_params
        self.__extra_cookie_params = extra_cookie_params

    def __call__(self, r: PreparedRequest) -> PreparedRequest:
        if self.__query_params:
            r.prepare_url(r.url, self.__query_params)
        r.headers.pop("Cookie", None)
        new_cookie = create_cookie(  # type: ignore
            name=self.__cookie_name,
            value=self.__cookie_provider(),
            **self.__extra_cookie_params
        )
        r._cookies.set_cookie(new_cookie)  # type: ignore
        r.prepare_cookies(r._cookies)  # type: ignore
        return r


class CookieAuthScheme(AuthScheme):
    __cookie_name: str
    __cookie_provider: PyLegendCallable[[], str]
    __query_params: PyLegendOptional[PyLegendDict[str, str]]

    def __init__(  # type: ignore
            self,
            cookie_name: str,
            cookie_provider: PyLegendCallable[[], str],
            query_params: PyLegendOptional[PyLegendDict[str, str]] = None,
            **extra_cookie_params
    ) -> None:
        self.__cookie_name = cookie_name
        self.__cookie_provider = cookie_provider
        self.__query_params = query_params
        self.__extra_cookie_params = extra_cookie_params

    def get_auth_base(self) -> PyLegendOptional[AuthBase]:
        return CookieAuth(self.__cookie_name, self.__cookie_provider, self.__query_params, **self.__extra_cookie_params)



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class SqlToStringFormat:
    pretty: bool
    indent_count: int

    def __init__(self, pretty: bool = True, indent_count: int = 0) -> None:
        self.pretty = pretty
        self.indent_count = indent_count

    def push_indent(self) -> "SqlToStringFormat":
        return SqlToStringFormat(self.pretty, self.indent_count + 1)

    def separator(self, cnt: int = 0) -> str:
        if self.pretty:
            return "\n" + "".join(["    " for _ in range(self.indent_count + cnt)])
        else:
            return " "


class SqlToStringConfig:
    format: SqlToStringFormat

    def __init__(
        self,
        format_: SqlToStringFormat
    ) -> None:
        self.format = format_

    def push_indent(self) -> "SqlToStringConfig":
        return SqlToStringConfig(self.format.push_indent())
